import cooler
import numpy as np
import pandas as pd
from scipy.sparse import load_npz, csr_matrix, save_npz, triu
from scipy.stats import rankdata


def _t_score(T, T2, tot):
    t = T2 - T.power(2)
    t.data = np.sqrt((tot - 1) / t.data)
    t = T.multiply(t)
    t.data[np.isnan(t.data)] = 0
    return t


def compute_t(group_prefix, tot=None):
    cool_t = cooler.Cooler(f'{group_prefix}.E.cool')
    cool_t2 = cooler.Cooler(f'{group_prefix}.E2.cool')
    if not tot:
        tot = cool_t.info['group_n_cells']
    for chrom in cool_t.chromnames:
        T = triu(cool_t.matrix(balance=False, sparse=True).fetch(chrom))
        T2 = triu(cool_t2.matrix(balance=False, sparse=True).fetch(chrom))
        t = _t_score(T, T2, tot)
        save_npz(f'{group_prefix}_{chrom}.tglobal.npz', t)

    cool_t = cooler.Cooler(f'{group_prefix}.T.cool')
    cool_t2 = cooler.Cooler(f'{group_prefix}.T2.cool')
    for chrom in cool_t.chromnames:
        T = triu(cool_t.matrix(balance=False, sparse=True).fetch(chrom))
        T2 = triu(cool_t2.matrix(balance=False, sparse=True).fetch(chrom))
        t = _t_score(T, T2, tot)
        save_npz(f'{group_prefix}_{chrom}.tlocal.npz', t)
    return tot


def permute_fdr(chrom_size_path,
                black_list_path,
                shuffle_group_prefix,
                real_group_prefix,
                res=10000,
                pad=7,
                min_dist=5,
                max_dist=500):
    chrom_size_series = pd.read_csv(chrom_size_path,
                                    sep='\t',
                                    index_col=0,
                                    header=None,
                                    squeeze=True)
    chrom_size = (chrom_size_series.values // res).astype(int) + 1
    chroms = chrom_size_series.index
    covfilter = []
    cumsize = [[0] for _ in range(min_dist, max_dist)]
    bkl = pd.read_csv(black_list_path, sep='\t', header=None, index_col=None)
    for k, chrom in enumerate(chroms):
        cov = np.zeros(chrom_size[k])
        for xx, yy in bkl.loc[bkl[0] == chrom, [1, 2]].values // res:
            cov[max([xx - pad, 0]):min([len(cov), yy + pad + 1])] = 1
        tmp = []
        row, col = np.diag_indices(chrom_size[k])
        for i in range(min_dist, max_dist):
            tmp.append(np.logical_or(cov[row[:-i]], cov[col[i:]]))
            cumsize[i - min_dist].append(cumsize[i - min_dist][-1] +
                                         np.sum(~tmp[-1]))
        covfilter.append(tmp)

    for bktype in ['local', 'global']:
        t = [[] for _ in range(min_dist, max_dist)]
        tnull = [[] for _ in range(min_dist, max_dist)]
        for k, chrom in enumerate(chroms):
            tmp = load_npz(
                f'{shuffle_group_prefix}_{chrom}.t{bktype}.npz').toarray()
            for i in range(min_dist, max_dist):
                tmp0 = tmp.diagonal(i).copy()[~covfilter[k][i - min_dist]]
                tnull[i - min_dist].append(tmp0)
            tmp = load_npz(
                f'{real_group_prefix}_{chrom}.t{bktype}.npz').toarray()
            for i in range(min_dist, max_dist):
                tmp0 = tmp.diagonal(i).copy()[~covfilter[k][i - min_dist]]
                t[i - min_dist].append(tmp0)

        fdr = []
        for i in range(len(t)):
            tmp1 = -np.concatenate(t[i])
            tmp2 = -np.concatenate(tnull[i])
            t1 = rankdata(np.concatenate((tmp1, tmp2)),
                          method='max')[:len(tmp1)]
            t2 = rankdata(tmp1, method='max')
            fdr.append((t1 - t2) / t2)

        for k, chrom in enumerate(chroms):
            ngene = chrom_size[k]
            tmp = np.zeros((ngene, ngene))
            row, col = np.diag_indices(ngene)
            for i in range(min_dist, max_dist):
                idx = (row[:-i][~covfilter[k][i - min_dist]],
                       col[i:][~covfilter[k][i - min_dist]])
                tmp[row[:-i], col[i:]] = 1
                tmp[idx] = fdr[i - min_dist][(cumsize[i - min_dist][k]):(
                    cumsize[i - min_dist][k + 1])]
            tmp = csr_matrix(tmp)
            save_npz(f'{shuffle_group_prefix}_{chrom}.permutefdr{bktype}.npz',
                     tmp)
    return


def update_fdr_qval(chrom_size_path,
                    real_group_prefix,
                    shuffle_group_prefix,
                    res=10000,
                    min_dist=5,
                    max_dist=500):
    chrom_size_series = pd.read_csv(chrom_size_path,
                                    sep='\t',
                                    index_col=0,
                                    header=None,
                                    squeeze=True)
    data: pd.DataFrame = pd.read_hdf(f'{real_group_prefix}.totalloop_info.hdf',
                                     key='data')
    data['global_qval'] = 1
    data['local_qval'] = 1
    data = data.loc[((data['distance'] // res) > min_dist)
                    & ((data['distance'] // res) < max_dist)
                    & data['bkfilter']]
    for chrom in chrom_size_series.index:
        tmpfilter = (data['chrom'] == chrom)
        tmp = data.loc[tmpfilter, ['x1', 'y1']].values // res
        coord = (tmp[:, 0], tmp[:, 1])
        tmp = load_npz(f'{shuffle_group_prefix}_{chrom}.permutefdrlocal.npz')
        data.loc[tmpfilter, 'local_qval'] = tmp[coord].A.ravel()
        tmp = load_npz(f'{shuffle_group_prefix}_{chrom}.permutefdrglobal.npz')
        data.loc[tmpfilter, 'global_qval'] = tmp[coord].A.ravel()

    data.to_hdf(f'{real_group_prefix}.totalloop_info.hdf',
                key='data',
                format='table')
    return data



"""
loop_bkg.py corresponding to draft/loop_bkg_cell, which calculates E and T from cell imputed matrix
merge_cell_to_group.py corresponding to loop_sumcell_chr,
which aggregates the cell E, T to cluster level E, E2, T, T2, O,
"""

from .snakemake import call_loop



import pandas as pd
from collections import defaultdict
import pybedtools
from functools import lru_cache


@lru_cache()
def prepare_2d_blacklist_dict(blacklist_bedpe, resolution=10000):
    # read blacklist bed file, turn the region into bad pixel idx dict with resolution
    blacklist_bedpe_df = pd.read_csv(blacklist_bedpe, sep='\t', header=None)

    # turn region into region pixel idx
    blacklist_bedpe_df[1] //= resolution
    blacklist_bedpe_df[2] //= resolution
    # in case the region is smaller than resolution, add one so at least one pixel is bad
    blacklist_bedpe_df.loc[2, (blacklist_bedpe_df[2] -
                               blacklist_bedpe_df[1]) < 1] += 1
    blacklist_bedpe_df[4] //= resolution
    blacklist_bedpe_df[5] //= resolution
    # in case the region is smaller than resolution, add one so at least one pixel is bad
    blacklist_bedpe_df.loc[5, (blacklist_bedpe_df[5] -
                               blacklist_bedpe_df[4]) < 1] += 1

    chrom_pair_bad_points = defaultdict(set)
    for idx, row in blacklist_bedpe_df.iterrows():
        for i in range(row[1], row[2]):
            for j in range(row[4], row[5]):
                chrom_pair_bad_points[row[0], row[3]].add((i, j))
                # in case contact is not ordered as blacklist
                chrom_pair_bad_points[row[3], row[0]].add((j, i))

    # return a dict, key is chrom pair
    return chrom_pair_bad_points


def _is_2d_blacklist(row, blacklist_2d):
    chrom1, chrom2, pos1, pos2 = row
    judge = (pos1, pos2) in blacklist_2d[(chrom1, chrom2)]
    return judge


def filter_contacts(contact_path,
                    chrom_size_path=None,
                    blacklist_1d_path=None,
                    blacklist_2d_path=None,
                    remove_duplicates=True,
                    resolution_2d=10000,
                    chrom1=1,
                    pos1=2,
                    chrom2=5,
                    pos2=6):
    try:
        contacts = pd.read_csv(contact_path,
                               header=None,
                               sep='\t',
                               dtype={
                                   chrom1: str,
                                   pos1: int,
                                   chrom2: str,
                                   pos2: int
                               })
    except Exception as e:
        print(f'Got error when opening {contact_path}')
        raise e
    print(f"{contact_path.split('/')[-1]}: {contacts.shape[0]} input contacts.")

    if remove_duplicates:
        # remove duplicates
        contacts.drop_duplicates(subset=[chrom1, pos1, chrom2, pos2], inplace=True)

    chroms = pd.read_csv(chrom_size_path, sep='\t', index_col=0, header=None).index
    # remove additional chroms not exist in chrom_size_path
    contacts = contacts[contacts[chrom1].isin(chroms) & contacts[chrom2].isin(chroms)].copy()

    if blacklist_1d_path is not None:
        blacklist_bed_df = pd.read_csv(blacklist_1d_path, sep='\t', index_col=None, header=None)
        blacklist_bed_df = blacklist_bed_df[blacklist_bed_df.iloc[:, 0].isin(chroms)].copy()
        blacklist_bed = pybedtools.BedTool.from_dataframe(blacklist_bed_df).sort(g=chrom_size_path)

        # determine blacklist 1d (either side overlap with 1D blacklist)
        left_bed_df = contacts[[chrom1, pos1, pos1]].reset_index().iloc[:, [1, 2, 3, 0]]
        right_bed_df = contacts[[chrom2, pos2, pos2]].reset_index().iloc[:, [1, 2, 3, 0]]
        left_bed_df.columns = ['chrom', 'start', 'end', 'id']
        right_bed_df.columns = ['chrom', 'start', 'end', 'id']
        contact_bed_df = pd.concat([left_bed_df, right_bed_df])
        contact_bed = pybedtools.BedTool.from_dataframe(contact_bed_df).sort(g=chrom_size_path)
        # collect contact ids with either side overlap with blacklist
        bad_contacts = contact_bed.intersect(blacklist_bed, wa=True, u=True).to_dataframe()
        if bad_contacts.shape[0] > 0:
            bad_contacts = bad_contacts['name'].unique()
        else:
            # no bad contacts
            bad_contacts = set()
        # remove bad contacts
        contacts = contacts[~contacts.index.isin(bad_contacts)].copy()

        pybedtools.cleanup()

    if blacklist_2d_path is not None:
        chrom_2d_blacklist = prepare_2d_blacklist_dict(blacklist_2d_path, resolution=resolution_2d)
        contacts_idx = contacts[[chrom1, chrom2, pos1, pos2]].copy()
        # turn contact location into bin idx with resolution
        contacts_idx[pos1] //= resolution_2d
        contacts_idx[pos2] //= resolution_2d

        # determine blacklist 2d (both side overlap with 2D blacklist)
        is_blacklist_2d = contacts_idx.apply(_is_2d_blacklist,
                                             blacklist_2d=chrom_2d_blacklist,
                                             axis=1)
        contacts = contacts[~is_blacklist_2d].copy()

    print(f"{contact_path.split('/')[-1]}: {contacts.shape[0]} filtered contacts in scool.")
    return contacts



# command time python /gale/ddn/snm3C/humanPFC/code/domain_insulation_cell.py --indir /gale/ddn/snm3C/humanPFC/smoothed_matrix/${res0}b_resolution/ --cell ${sample} --chrom ${c} --mode pad2_std1_rp0.5_sqrtvc --w 10

import h5py
import argparse
import numpy as np
from scipy.sparse import csr_matrix

def domain_insulation_cell(indir, cell, chrom, mode, w=10):

	if chrom[:3]=='chr':
		c = chrom
	else:
		c = 'chr' + chrom
	with h5py.File(f'{indir}{c}/{cell}_{c}_{mode}.hdf5', 'r') as f:
		g = f['Matrix']
		A = csr_matrix((g['data'][()], g['indices'][()], g['indptr'][()]), g.attrs['shape'])
	score = np.ones(A.shape[0])
	for i in range(1,A.shape[0]):
		if i<w:
			intra = (A[:i, :i].sum() + A[i:(i+w), i:(i+w)].sum()) / (i*(i+1)/2 + (w*(w+1)/2))
			inter = A[:i, i:(i+w)].sum() / (i * (i+w))
		else:
			intra = (A[(i-w):i, (i-w):i].sum() + A[i:(i+w), i:(i+w)].sum()) / (w*(w+1))
			inter = A[(i-w):i, i:(i+w)].sum() / (w*w)
		score[i] = inter / (inter + intra)
	np.save(f'{indir}{c}/{cell}_{c}_{mode}.w{w}.ins.npy', score)
	return

'''
parser = argparse.ArgumentParser()
parser.add_argument('--indir', type=str, default=None, help='Directory of imputed matrices end with /')
parser.add_argument('--cell', type=str, default=None, help='Full path of a file containing a list of cell identifiers to be concatenate')
parser.add_argument('--chrom', type=str, default=None, help='Chromosome to impute')
parser.add_argument('--mode', type=str, default=None, help='Suffix of imputed matrix file names')
parser.add_argument('--w', type=int, default=10, help='Window size for insulation score')
opt = parser.parse_args()

domain_insulation_cell(opt.indir, opt.cell, opt.chrom, opt.mode, opt.w)
'''



import sys
import numpy as np
import pandas as pd

chrom = [f'chr{i + 1}' for i in range(22)]
loop1 = pd.read_csv(sys.argv[1], sep='\t', header=None, index_col=None).sort_values(by=[0, 1, 4])
loop2 = pd.read_csv(sys.argv[2], sep='\t', header=None, index_col=None).sort_values(by=[0, 1, 4])
loop1 = loop1[loop1[0].isin(chrom)]
loop2 = loop2[loop2[0].isin(chrom)]
res = loop1.iloc[0, 2] - loop1.iloc[0, 1]
tot = 0
dist = 20000
loop0 = []
for c in chrom:
    tmp1 = loop1.loc[loop1[0] == c, [1, 4]].values
    tmp2 = loop2.loc[loop2[0] == c, [1, 4]].values
    j, tmp = 0, 0
    for i in range(len(tmp1)):
        while j <= (len(tmp2) - 1) and tmp2[j, 0] < (tmp1[i, 0] - dist):
            j += 1
        k = j
        flag = 1
        while k <= (len(tmp2) - 1) and tmp2[k, 0] <= (tmp1[i, 0] + dist):
            if (tmp1[i, 1] - dist) <= tmp2[k, 1] <= (tmp1[i, 1] + dist):
                tmp += 1
                flag = 0
                break
            k += 1
        if flag:
            loop0.append([c, tmp1[i, 0], tmp1[i, 0] + res, c, tmp1[i, 1], tmp1[i, 1] + res])
    tot += tmp

print(tot, loop1.shape[0], tot / loop1.shape[0])
if len(sys.argv) == 3:
    np.savetxt('.'.join(sys.argv[1].split('.')[:-1]) + '.specific.bedpe',
               loop0, fmt='%s', delimiter='\t')



import pathlib
import subprocess

import pandas as pd
import inspect

import schicluster
from .loop_calling import filter_loops
from .merge_raw_matrix import make_raw_matrix_cell_table
from .shuffle_fdr import compute_t, permute_fdr, update_fdr_qval

PACKAGE_DIR = pathlib.Path(schicluster.__path__[0])

with open(PACKAGE_DIR / 'loop/generate_matrix_chunk.Snakefile') as tmp:
    GENERATE_MATRIX_CHUNK_TEMPLATE = tmp.read()
with open(PACKAGE_DIR / 'loop/generate_matrix_group.Snakefile') as tmp:
    GENERATE_MATRIX_SCOOL_TEMPLATE = tmp.read()


def prepare_dir(output_dir, chunk_df, dist, cap, pad, gap, resolution,
                min_cutoff, chrom_size_path, keep_cell_matrix, log_e_str, shuffle):
    output_dir.mkdir(exist_ok=True)
    cell_table_path = str((output_dir / 'cell_table.csv').absolute())
    chunk_df[['cell_url']].to_csv(cell_table_path)
    if shuffle:
        shuffle_str = '--shuffle'
    else:
        shuffle_str = ''
    parameters = dict(dist=dist,
                      cap=cap,
                      pad=pad,
                      gap=gap,
                      resolution=resolution,
                      min_cutoff=min_cutoff,
                      cell_table_path=f'"{cell_table_path}"',
                      chrom_size_path=f'"{chrom_size_path}"',
                      keep_cell_matrix=keep_cell_matrix,
                      log_e_str=f'"{log_e_str}"',
                      shuffle=shuffle,
                      shuffle_str=f'"{shuffle_str}"')
    parameters_str = '\n'.join(f'{k} = {v}'
                               for k, v in parameters.items())

    with open(output_dir / 'Snakefile', 'w') as f:
        f.write(parameters_str + GENERATE_MATRIX_CHUNK_TEMPLATE)
    return


def prepare_loop_snakemake(cell_table_path,
                           output_dir,
                           chrom_size_path,
                           chunk_size=100,
                           dist=5050000,
                           cap=5,
                           pad=5,
                           gap=2,
                           resolution=10000,
                           min_cutoff=1e-6,
                           keep_cell_matrix=False,
                           cpu_per_job=10,
                           log_e=True,
                           shuffle=False,
                           raw_resolution_str=None,
                           downsample_shuffle=None):
    _cell_table_path = str(cell_table_path)
    sep = '\t' if _cell_table_path.endswith('tsv') else ','
    cell_table = pd.read_csv(cell_table_path, index_col=0, sep=sep, header=None,
                             names=['cell_id', 'cell_url', 'cell_group'])
    if shuffle and (downsample_shuffle is not None):
        # for shuffle background, downsample to downsample_shuffle to save time
        if cell_table.shape[0] > downsample_shuffle:
            cell_table = cell_table.sample(downsample_shuffle)
    output_dir = pathlib.Path(output_dir).absolute()
    output_dir.mkdir(exist_ok=True)

    # a single dir for raw matrix
    if raw_resolution_str == '10K':
        raw_resolution = 10000
    elif raw_resolution_str is None:
        raw_resolution = None
    else:
        raise NotImplementedError
    if (raw_resolution is not None) and (not shuffle):
        cell_table_raw = make_raw_matrix_cell_table(cell_table, raw_resolution_str)
        raw_dir = output_dir / 'raw'
        raw_dir.mkdir(exist_ok=True)
        raw_table_path = raw_dir / 'cell_table.tsv'
        cell_table_raw.to_csv(raw_table_path, sep='\t', header=None)
        merge_raw_cmd = f'hic-internal merge-raw-scool ' \
                        f'--chrom_size_path {chrom_size_path} ' \
                        f'--resolution {raw_resolution} ' \
                        f'--cell_table_path {raw_table_path} ' \
                        f'--output_dir {raw_dir} ' \
                        f'--cpu {cpu_per_job}'
    else:
        merge_raw_cmd = ''

    if log_e:
        log_e_str = '--log_e'
    else:
        log_e_str = ''
    chunk_parameters = dict(
        dist=dist,
        cap=cap,
        pad=pad,
        gap=gap,
        resolution=resolution,
        min_cutoff=min_cutoff,
        chrom_size_path=chrom_size_path,
        keep_cell_matrix=keep_cell_matrix,
        log_e_str=log_e_str,
        shuffle=shuffle
    )

    total_chunk_dirs = []
    group_chunks = {}
    for group, group_df in cell_table.groupby('cell_group'):
        group_chunks[group] = []
        if group_df.shape[0] <= chunk_size:
            this_dir = output_dir / f'{group}_chunk0'
            prepare_dir(this_dir, group_df, **chunk_parameters)
            total_chunk_dirs.append(this_dir)
            group_chunks[group].append(this_dir)
        else:
            group_df['chunk'] = [i // chunk_size for i in range(group_df.shape[0])]
            for chunk, chunk_df in group_df.groupby('chunk'):
                this_dir = output_dir / f'{group}_chunk{chunk}'
                prepare_dir(this_dir, chunk_df, **chunk_parameters)
                total_chunk_dirs.append(this_dir)
                group_chunks[group].append(this_dir)

    with open(output_dir / 'snakemake_cmd_step1.txt', 'w') as f:
        f.write(merge_raw_cmd + '\n')
        for chunk_dir in total_chunk_dirs:
            cmd = f'snakemake -d {chunk_dir} --snakefile {chunk_dir}/Snakefile -j {cpu_per_job}'
            f.write(cmd + '\n')

    # prepare the second step that merge cell chunks into groups
    scool_parameters = dict(
        output_dir=f'"{output_dir}"',
        chrom_size_path=f'"{chrom_size_path}"',
        resolution=resolution,
        shuffle=shuffle
    )
    parameters_str = '\n'.join(f'{k} = {v}'
                               for k, v in scool_parameters.items())
    with open(output_dir / 'Snakefile', 'w') as f:
        f.write(parameters_str + GENERATE_MATRIX_SCOOL_TEMPLATE)
    with open(output_dir / 'snakemake_cmd_step2.txt', 'w') as f:
        cmd = f'snakemake -d {output_dir} --snakefile {output_dir}/Snakefile -j {cpu_per_job}'
        f.write(cmd + '\n')
    return


def check_chunk_dir_finish(output_dir):
    output_dir = pathlib.Path(output_dir)
    flag_path = output_dir / 'chunk_finished'
    if flag_path.exists():
        return

    path_not_finished = []
    for chunk_dir in output_dir.glob('*_chunk*'):
        if not (chunk_dir / 'finish').exists():
            path_not_finished.append(str(chunk_dir))
    if len(path_not_finished) > 0:
        path_str = '\n'.join(path_not_finished)
        raise ValueError('These chunk dirs have not finish successfully:\n' + path_str)
    else:
        with open(flag_path, 'w') as _:
            pass
    return


def _run_snakemake(output_dir):
    output_dir = pathlib.Path(output_dir).absolute()
    step1 = f'{output_dir}/snakemake_cmd_step1.txt'
    step2 = f'{output_dir}/snakemake_cmd_step2.txt'
    subprocess.run(['sh', step1], check=True)
    subprocess.run(['sh', step2], check=True)


def call_loop(cell_table_path,
              output_dir,
              chrom_size_path,
              shuffle=True,
              chunk_size=200,
              dist=5050000,
              cap=5,
              pad=5,
              gap=2,
              resolution=10000,
              min_cutoff=1e-6,
              keep_cell_matrix=False,
              cpu_per_job=10,
              log_e=True,
              raw_resolution_str=None,
              downsample_shuffle=None,
              black_list_path=None,
              fdr_pad=7,
              fdr_min_dist=5,
              fdr_max_dist=500,
              fdr_thres=0.1,
              dist_thres=20000,
              size_thres=1,
              cleanup=True):
    if shuffle and (black_list_path is None):
        raise ValueError('Please provide black_list_path when shuffle=True')

    pathlib.Path(output_dir).mkdir(exist_ok=True)

    # deal with cell table path, if the path has two col, add one group cal internally
    _cell_table_path = str(cell_table_path)
    sep = '\t' if _cell_table_path.endswith('tsv') else ','
    cell_table = pd.read_csv(cell_table_path, index_col=0, sep=sep, header=None)
    cell_table.index.name = 'cell_id'
    if cell_table.shape[1] == 1:
        cell_table.columns = ['cell_url']
        cell_table['cell_group'] = 'group'
    elif cell_table.shape[1] == 2:
        cell_table.columns = ['cell_url', 'cell_group']
    else:
        raise ValueError(f'Expect cell_table_path to be '
                         f'two columns (cell_id, cell_url) or '
                         f'three columns (cell_id, cell_url, cell_group), '
                         f'got {cell_table.shape[1]} columns.')
    cell_table_path = f'{output_dir}/cell_table.csv'
    cell_table.to_csv(cell_table_path, header=None)
    groups = cell_table['cell_group'].unique()

    kwargs = locals()
    shuffle = kwargs.pop('shuffle')

    # prepare snakemake and execute
    real_dir = None
    shuffle_dir = None
    _use_kwargs = {k: v
                   for k, v in kwargs.items()
                   if k in inspect.signature(prepare_loop_snakemake).parameters}
    if shuffle:
        output_dir = _use_kwargs.pop('output_dir')
        real_dir = output_dir
        shuffle_dir = f'{output_dir}/shuffle'
        pathlib.Path(shuffle_dir).mkdir(exist_ok=True, parents=True)
        prepare_loop_snakemake(shuffle=False, output_dir=real_dir, **_use_kwargs)
        prepare_loop_snakemake(shuffle=True, output_dir=shuffle_dir, **_use_kwargs)
        _run_snakemake(real_dir)
        _run_snakemake(shuffle_dir)
    else:
        # not shuffle, just use normal loop pipeline
        prepare_loop_snakemake(shuffle=False, **_use_kwargs)
        output_dir = _use_kwargs.pop('output_dir')
        _run_snakemake(output_dir)

    # final call loop if shuffle
    if shuffle:
        for group in groups:
            real_group_prefix = f'{real_dir}/{group}/{group}'
            shuffle_group_prefix = f'{shuffle_dir}/{group}/{group}'

            tot = compute_t(real_group_prefix)
            _ = compute_t(shuffle_group_prefix, tot)

            permute_fdr(chrom_size_path=chrom_size_path,
                        black_list_path=black_list_path,
                        shuffle_group_prefix=shuffle_group_prefix,
                        real_group_prefix=real_group_prefix,
                        res=resolution,
                        pad=fdr_pad,
                        min_dist=fdr_min_dist,
                        max_dist=fdr_max_dist)

            total_loops = update_fdr_qval(chrom_size_path,
                                          real_group_prefix,
                                          shuffle_group_prefix,
                                          res=resolution,
                                          min_dist=fdr_min_dist,
                                          max_dist=fdr_max_dist)

            # redo filter loops because FDR changed
            filter_loops(total_loops,
                         output_prefix=real_group_prefix,
                         fdr_thres=fdr_thres,
                         resolution=resolution,
                         dist_thres=dist_thres,
                         size_thres=size_thres)

    if cleanup:
        # subprocess.run(f'rm -rf {output_dir}/shuffle', shell=True)
        subprocess.run(f'rm -rf {output_dir}/*/*global.npz', shell=True)
        subprocess.run(f'rm -rf {output_dir}/*/*local.npz', shell=True)

    with open(f'{output_dir}/Success', 'w') as f:
        f.write('42')
    return



import re
import subprocess
import warnings
from collections import Counter, defaultdict
from concurrent.futures import ProcessPoolExecutor, as_completed

import cooler
import pandas as pd
import pandas.errors
from cooler import create_scool

from .remove_blacklist import filter_contacts
from .utilities import get_chrom_offsets


def generate_scool_batch_data(cell_path_dict,
                              resolution,
                              chrom_offset,
                              chrom_size_path,
                              blacklist_1d_path,
                              blacklist_2d_path,
                              remove_duplicates,
                              blacklist_resolution,
                              output_path,
                              chr1=1,
                              chr2=5,
                              pos1=2,
                              pos2=6,
                              min_pos_dist=2500):
    def single_cell_pixel(_path):
        try:
            contacts = filter_contacts(_path,
                                       chrom_size_path=chrom_size_path,
                                       blacklist_1d_path=blacklist_1d_path,
                                       blacklist_2d_path=blacklist_2d_path,
                                       remove_duplicates=remove_duplicates,
                                       resolution_2d=blacklist_resolution,
                                       chrom1=chr1,
                                       pos1=pos1,
                                       chrom2=chr2,
                                       pos2=pos2)
        except pandas.errors.EmptyDataError:
            # empty contacts file
            return pd.DataFrame([], columns=['bin1_id', 'bin2_id', 'count'])
        pos_dist = (contacts[pos1] - contacts[pos2]).abs()

        # filter
        contacts = contacts[contacts[chr1].isin(chrom_offset)
                            & contacts[chr2].isin(chrom_offset)
                            & (contacts[pos1] > 0)
                            & (contacts[pos2] > 0)]
        # apply min_pos_dist filter on cis contact only
        contacts = contacts[(pos_dist > min_pos_dist) |  # filter all close contact
                            (contacts[chr1] != contacts[chr2])]  # but keep all trans

        # calculate pixel
        bin1_id = contacts[chr1].map(chrom_offset) + (contacts[pos1] - 1) // resolution
        bin2_id = contacts[chr2].map(chrom_offset) + (contacts[pos2] - 1) // resolution
        counter = Counter()
        for x, y in zip(bin1_id, bin2_id):
            if x > y:
                x, y = y, x
            counter[(x, y)] += 1
        pixel_df = pd.Series(counter).reset_index()
        pixel_df.columns = pd.Index(['bin1_id', 'bin2_id', 'count'])
        pixel_df = pixel_df.sort_values(['bin1_id',
                                         'bin2_id']).reset_index(drop=True)
        return pixel_df

    with warnings.catch_warnings():
        # ignore the
        warnings.simplefilter("ignore")
        with pd.HDFStore(output_path) as hdf:
            for cell_id, path in cell_path_dict.items():
                hdf[cell_id] = single_cell_pixel(path)
    return


def generate_scool_single_resolution(cell_path_dict,
                                     chrom_size_path,
                                     resolution,
                                     output_path,
                                     blacklist_1d_path,
                                     blacklist_2d_path,
                                     remove_duplicates,
                                     blacklist_resolution,
                                     chr1=1,
                                     chr2=5,
                                     pos1=2,
                                     pos2=6,
                                     min_pos_dist=2500,
                                     batch_n=20,
                                     cpu=1):
    # parse chromosome sizes, prepare bin_df
    chrom_sizes = pd.read_csv(
        chrom_size_path,
        sep='\t',
        index_col=0,
        header=None, squeeze=True)
    bins_df = cooler.binnify(chrom_sizes, resolution)
    chrom_offset = get_chrom_offsets(bins_df)

    chunk_dicts = defaultdict(dict)
    for i, (cell, path) in enumerate(cell_path_dict.items()):
        chunk_dicts[i // batch_n][cell] = path

    with ProcessPoolExecutor(cpu) as exe:
        futures = {}
        for batch, cell_path_dict in chunk_dicts.items():
            batch_output = output_path + f'_{batch}'
            f = exe.submit(generate_scool_batch_data,
                           cell_path_dict=cell_path_dict,
                           resolution=resolution,
                           chrom_offset=chrom_offset,
                           output_path=batch_output,
                           chr1=chr1, chr2=chr2,
                           pos1=pos1, pos2=pos2,
                           min_pos_dist=min_pos_dist,
                           chrom_size_path=chrom_size_path,
                           blacklist_1d_path=blacklist_1d_path,
                           blacklist_2d_path=blacklist_2d_path,
                           remove_duplicates=remove_duplicates,
                           blacklist_resolution=blacklist_resolution)
            futures[f] = batch_output

        for future in as_completed(futures):
            # batch finished
            batch_output = futures[future]
            future.result()

            # dump batch result into cool
            cell_pixel_dict = {}
            with pd.HDFStore(batch_output, mode='r') as hdf:
                for cell_id in hdf.keys():
                    cell_id = cell_id[1:]  # remove '/'
                    cell_pixel_dict[cell_id] = hdf[cell_id]
            create_scool(output_path,
                         bins=bins_df,
                         cell_name_pixels_dict=cell_pixel_dict,
                         ordered=True,
                         mode='a')
            subprocess.run(['rm', '-f', batch_output], check=True)
    return


def generate_scool(contacts_table,
                   output_prefix,
                   chrom_size_path,
                   resolutions,
                   blacklist_1d_path=None,
                   blacklist_2d_path=None,
                   blacklist_resolution=10000,
                   remove_duplicates=True,
                   chr1=1,
                   chr2=5,
                   pos1=2,
                   pos2=6,
                   min_pos_dist=2500,
                   cpu=1,
                   batch_n=50):
    """
    Generate single-resolution cool files from single-cell contact files recorded in contacts_table

    Parameters
    ----------
    contacts_table
        tab-separated table containing tow columns, 1) cell id, 2) cell contact file path (juicer-pre format)
        No header
    output_prefix
        Output prefix of the cool files. Output path will be {output_prefix}.{resolution_str}.cool
    chrom_size_path
        Path to the chromosome size file, this file should be in UCSC chrom.sizes format. We will use this file as
        the reference to create matrix. It is recommended to remove small contigs or chrM from this file.
    resolutions
        Resolutions to generate the matrix. Each resolution will be stored in a separate file.
    blacklist_1d_path
        Path to blacklist region BED file, such as ENCODE blacklist.
        Either side of the contact overlapping with a blacklist region will be removed.
    blacklist_2d_path
        Path to blacklist region pair BEDPE file.
        Both side of the contact overlapping with the same blacklist region pair will be removed.
    blacklist_resolution
        Resolution in bps when consider the 2D blacklist region pairs.
    remove_duplicates
        If true, will remove duplicated contacts based on [chr1, pos1, chr2, pos2] values
    chr1
        0 based index of chr1 column.
    chr2
        0 based index of chr2 column.
    pos1
        0 based index of pos1 column.
    pos2
        0 based index of pos2 column.
    min_pos_dist
        Minimum distance for a fragment to be considered.
    cpu
        number of cpus to parallel.
    batch_n
        number of cells to deal with in each cpu process.

    Returns
    -------

    """
    # read cell paths
    cell_path_dict = pd.read_csv(contacts_table,
                                 header=None,
                                 index_col=0,
                                 squeeze=True,
                                 sep='\t').to_dict()
    print(f'{len(cell_path_dict)} cells to process.')

    for resolution in resolutions:
        resolution_str = str(resolution)
        resolution_str = re.sub(r'000000$', 'M', resolution_str)
        resolution_str = re.sub(r'000$', 'K', resolution_str)
        output_path = f'{output_prefix}.{resolution_str}.scool'

        print('Generating', output_path)
        generate_scool_single_resolution(cell_path_dict=cell_path_dict,
                                         chrom_size_path=chrom_size_path,
                                         resolution=resolution,
                                         output_path=output_path,
                                         batch_n=batch_n,
                                         cpu=cpu,
                                         chr1=chr1, chr2=chr2,
                                         pos1=pos1, pos2=pos2,
                                         min_pos_dist=min_pos_dist,
                                         blacklist_1d_path=blacklist_1d_path,
                                         blacklist_2d_path=blacklist_2d_path,
                                         remove_duplicates=remove_duplicates,
                                         blacklist_resolution=blacklist_resolution)
        print('Finished', output_path)
    return



# for c in `seq 1 22`; do awk -v c=$c '{printf("/gale/ddn/snm3C/humanPFC/smoothed_matrix/25kb_resolution/chr%s/%s_chr%s_pad2_std1_rp0.5_sqrtvc.w10.domain.bed\n",c,$1,c)}' celllist_long.txt > 25kb_resolution/filelist/domainlist_pad2_std1_rp0.5_sqrtvc_chr${c}.txt; echo $c; done
# for c in `seq 1 22`; do awk -v c=$c '{printf("/gale/ddn/snm3C/humanPFC/smoothed_matrix/25kb_resolution/chr%s/%s_chr%s_pad2_std1_rp0.5_sqrtvc.w10.ins.npy\n",c,$1,c)}' celllist_long.txt > 25kb_resolution/filelist/inslist_pad2_std1_rp0.5_sqrtvc_chr${c}.txt; echo $c; done

# command time python /gale/ddn/snm3C/humanPFC/code/domain_concatcell_chr.py --cell_list /gale/ddn/snm3C/humanPFC/smoothed_matrix/25kb_resolution/filelist/inslist_pad2_std1_rp0.5_sqrtvc_chr${c}.txt --outprefix /gale/ddn/snm3C/humanPFC/smoothed_matrix/${res0}b_resolution/merged/pad2_std1_rp0.5_sqrtvc_chr${c}.w10 --res 25000 --input_type insulation --ncpus 10
# command time python /gale/ddn/snm3C/humanPFC/code/domain_concatcell_chr.py --cell_list /gale/ddn/snm3C/humanPFC/smoothed_matrix/25kb_resolution/filelist/domainlist_pad2_std1_rp0.5_sqrtvc_chr${c}.txt --outprefix /gale/ddn/snm3C/humanPFC/smoothed_matrix/${res0}b_resolution/merged/pad2_std1_rp0.5_sqrtvc_chr${c}.w10 --res 25000 --input_type boundary --ncpus 10

import argparse
import numpy as np
import pandas as pd
from multiprocessing import Pool
from scipy.sparse import csr_matrix, save_npz

def load_insulation(i):
	try:
		data = np.load(celllist[i])
	except:
		print(celllist[i])
		data = 0
	return [i, data]

def load_boundary(i):
	try:
		tmp = pd.read_csv(celllist[i], sep='\t', header=None)
	except:
		print(celllist[i])
		data = 0
	else:
		data = np.zeros(int(tmp.iloc[-1,2] // rs) + 1)
		tmp = tmp[tmp[3]=='domain'][[1,2]].values // rs
		data[tmp[:,0]] += 1
		data[tmp[:,1]] += 1
	return [i, data]

def domain_concatcell_chr(cell_list, outprefix, res, input_type='insulation', ncpus=10):

	global celllist, rs
	celllist = np.loadtxt(cell_list, dtype=np.str)
	rs = res

	p = Pool(ncpus)
	if input_type=='insulation':
		result = p.map(load_insulation, np.arange(len(celllist)))
	elif input_type=='boundary':
		result = p.map(load_boundary, np.arange(len(celllist)))
	ins = np.zeros((len(celllist), len(result[0][1])))
	for i,x in result:
		if not isinstance(x, int):
			ins[i] = x.copy()
	if input_type=='insulation':
		np.save(f'{outprefix}.{input_type}.npy', ins)
	elif input_type=='boundary':
		save_npz(f'{outprefix}.{input_type}.npz', csr_matrix(ins))
	p.close()
	return

'''
parser = argparse.ArgumentParser()
parser.add_argument('--cell_list', type=str, default=None, help='Full path of a file containing the full path of all insulation npy or domain txt files to be concatenate')
parser.add_argument('--outprefix', type=str, default=None, help='Prefix of concatenated matrix including directory')
parser.add_argument('--res', type=int, default=None, help='Bin size as integer')
parser.add_argument('--input_type', type=str, default='insulation', help='Whether input files are insulation.npy or domain.txt') # insulation or boundary
parser.add_argument('--ncpus', type=int, default=10, help='# threads for parallelization')
opt = parser.parse_args()

domain_concatcell_chr(opt.cell_list, opt.outprefix, opt.res, opt.input_type, opt.ncpus)
'''



# command time python /gale/ddn/snm3C/humanPFC/code/impute_cell.py --indir /gale/raidix/rdx-5/zhoujt/projects/methylHiC/PFC_batch_merged/smoothed_matrix/1cell/${res0}b_resolution/chr${c}/ --outdir /gale/ddn/snm3C/humanPFC/smoothed_matrix/${res0}b_resolution/chr${c}/ --cell ${sample} --chrom ${c} --res ${res} --chrom_file /gale/netapp/home/zhoujt/genome/hg19/hg19.autosomal.chrom.sizes --mode pad2_std1_rp0.5_sqrtvc
# command time python /gale/ddn/snm3C/humanPFC/code/impute_cell.py --indir /gale/ddn/snm3C/humanPFC/cell_matrix/${res0}b_resolution/chr${c}/ --outdir /gale/ddn/snm3C/humanPFC/smoothed_matrix/${res0}b_resolution/chr${c}/ --cell ${sample} --chrom ${c} --res ${res} --chrom_file /gale/netapp/home/zhoujt/genome/hg19/hg19.autosomal.chrom.sizes --pad 2 --output_dist 10050000 --window_size 40000000 --step_size 10000000 --mode pad2_std1_rp0.5_ws40

import os
import time
import h5py
import cv2
cv2.useOptimized()
import argparse
import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix, save_npz, diags, eye
from scipy.sparse.linalg import norm
from scipy import ndimage as nd

def impute_cell(indir, outdir, cell, chrom, res, chrom_file, 
                logscale=False, pad=1, std=1, rp=0.5, tol=0.01, window_size=500000000, step_size=10000000,
                output_dist=500000000, output_format='hdf5', mode=None):

    def random_walk_cpu(P, rp, tol):
        if rp==1:
            return P
        ngene = P.shape[0]
        I = eye(ngene)
        Q = P.copy()
        start_time = time.time()
        for i in range(30):
            Q_new = P.dot(Q * (1 - rp) + rp * I)
            delta = norm(Q - Q_new)
            Q = Q_new.copy()
            sparsity = Q.nnz / ngene / ngene
            end_time = time.time()
            print('Iter', i+1, 'takes', end_time-start_time, 'seconds, loss', delta, 'sparsity', sparsity)
            if delta < tol:
                break
        return Q

    def output():
        if output_format=='npz':
            save_npz(f'{outdir}{cell}_{c}_{mode}.npz', E.astype(np.float32))
        else:
            f = h5py.File(f'{outdir}{cell}_{c}_{mode}.hdf5', 'w')
            g = f.create_group('Matrix')
            g.create_dataset('data', data=E.data, dtype='float32', compression='gzip')
            g.create_dataset('indices', data=E.indices, dtype=int, compression='gzip') 
            g.create_dataset('indptr', data=E.indptr, dtype=int, compression='gzip')
            g.attrs['shape'] = E.shape
            f.close()
        return

    if not os.path.exists(outdir):
        print('Output directory does not exist')
        return

    if chrom[:3]=='chr':
        c = chrom
    else:
        c = 'chr' + chrom
    if not mode:
        mode = f'pad{str(pad)}_std{str(std)}_rp{str(rp)}_sqrtvc'

    ws = window_size // res
    ss = step_size // res

    chromsize = pd.read_csv(chrom_file, sep='\t', header=None, index_col=0).to_dict()[1]

    start_time = time.time()
    ngene = int(chromsize[c] // res) + 1
    D = np.loadtxt(f'{indir}{cell}_{c}.txt')
    # to avoid bugs on chromosomes with 0/1 read
    if len(D)==0:
        E = eye(ngene).tocsr()
        output()
        return

    elif len(D.shape)==1:
        D = D.reshape(1,-1)

    A = csr_matrix((D[:, 2], (D[:, 0], D[:, 1])), shape = (ngene, ngene))
    if logscale:
        A.data = np.log2(A.data + 1)

    end_time = time.time()
    print('Loading takes', end_time-start_time, 'seconds')

    start_time = time.time()
    if pad > 0:
        A = cv2.GaussianBlur((A + A.T).astype(np.float32).toarray(), (pad*2+1, pad*2+1), std)
    else:
        A = (A + A.T).astype(np.float32)

    end_time = time.time()
    print('Convolution takes', end_time-start_time, 'seconds')

    start_time = time.time()
    # remove diagonal before rwr
    A = csr_matrix(A)
    A = A - diags(A.diagonal())
    if ws>=ngene or rp==1:
        B = A + diags((A.sum(axis=0).A.ravel()==0).astype(int))
        d = diags(1 / B.sum(axis=0).A.ravel())
        P = d.dot(B)
        E = random_walk_cpu(P, rp, tol)
        E = E.multiply(E > 0.01/ngene)
    else:
        idx = (np.repeat(np.arange(ws), ws), np.tile(np.arange(ws), ws))
        idxfilter = (np.abs(idx[1] - idx[0]) < (output_dist // res + 1))
        idx = (idx[0][idxfilter], idx[1][idxfilter])
        # first filter
        idxfilter = ((idx[0] + idx[1]) < (ws + ss))
        idx1 = (idx[0][idxfilter], idx[1][idxfilter])
        mask1 = csr_matrix((np.ones(len(idx1[0])), (idx1[0], idx1[1])), (ws, ws))
        # last filter
        idxfilter = ((idx[0] + idx[1]) >= ((ngene-ws) // ss * 2 + 1) * ss + 3 * ws - 2 * ngene)
        idx2 = (idx[0][idxfilter], idx[1][idxfilter])
        mask2 = csr_matrix((np.ones(len(idx2[0])), (idx2[0], idx2[1])), (ws, ws))
        # center filter
        idxfilter = np.logical_and((idx[0] + idx[1]) < (ws + ss), (idx[0] + idx[1]) >= (ws - ss))
        idx0 = (idx[0][idxfilter], idx[1][idxfilter])
        mask0 = csr_matrix((np.ones(len(idx0[0])), (idx0[0], idx0[1])), (ws, ws))
        start_time = time.time()
        E = csr_matrix(A.shape)
        for ll in [x for x in range(0, ngene - ws, ss)] + [ngene - ws]:
            B = A[ll:(ll+ws), ll:(ll+ws)]
            B = B + diags((B.sum(axis=0).A.ravel()==0).astype(int))
            d = diags(1 / B.sum(axis=0).A.ravel())
            P = d.dot(B)
            Etmp = random_walk_cpu(P, rp, tol)
            if ll==0:
                E[ll:(ll+ws), ll:(ll+ws)] += Etmp.multiply(mask1)
            elif ll==(ngene-ws):
                E[ll:(ll+ws), ll:(ll+ws)] += Etmp.multiply(mask2)
            else:
                E[ll:(ll+ws), ll:(ll+ws)] += Etmp.multiply(mask0)
            print('Window', ll)
        E = E.multiply(E > 0.01/ws)
    print('RWR takes', time.time() - start_time, 'seconds')

    start_time = time.time()
    E = E + E.T
    d = E.sum(axis=0).A.ravel()
    d[d==0] = 1
    b = diags(1 / np.sqrt(d))
    E = b.dot(E).dot(b)
    print('SQRTVC takes', time.time() - start_time, 'seconds')

    # longest distance filter mask
    start_time = time.time()
    idx = np.triu_indices(E.shape[0], 0)
    if (output_dist // res + 1) < ngene:
        idxfilter = ((idx[1] - idx[0]) < (output_dist // res + 1))
        idx = (idx[0][idxfilter], idx[1][idxfilter])
    mask = csr_matrix((np.ones(len(idx[0])), (idx[0], idx[1])), E.shape)
    E = E.tocsr().multiply(mask)
    print('Filter takes', time.time() - start_time, 'seconds')

    output()
    return

'''
parser = argparse.ArgumentParser()
parser.add_argument('--indir', type=str, default=None, help='Directory of the contact matrix')
parser.add_argument('--outdir', type=str, default=None, help='Output directory end with /')
parser.add_argument('--cell', type=str, default=None, help='Specific identifier of a cell')
parser.add_argument('--chrom', type=str, default=None, help='Chromosome to impute')
parser.add_argument('--res', type=int, default=None, help='Bin size as integer to generate contact matrix')
parser.add_argument('--chrom_file', type=str, default=None, help='Path to the chromosome size files containing all chromosomes to be analyzed')

parser.add_argument('--logscale', dest='logscale', action='store_true', help='To log transform raw count')
parser.set_defaults(logscale=False)
parser.add_argument('--pad', type=int, default=1, help='Gaussian kernal size')
parser.add_argument('--std', type=float, default=1, help='Gaussian kernal standard deviation')

parser.add_argument('--rp', type=float, default=0.5, help='Restart probability of RWR')
parser.add_argument('--tol', type=float, default=0.01, help='Convergence tolerance of RWR')
parser.add_argument('--window_size', type=int, default=500000000, help='Size of RWR sliding window')
parser.add_argument('--step_size', type=int, default=10000000, help='Step length of RWR sliding window')

parser.add_argument('--output_dist', type=int, default=500000000, help='Maximum distance threshold of contacts when writing output file')
parser.add_argument('--output_format', type=str, default='hdf5', help='Output file format (hdf5 or npz)')
parser.add_argument('--mode', type=str, default=None, help='Suffix of output file name')
opt = parser.parse_args()

impute_cell(opt.indir, opt.outdir, opt.cell, opt.chrom, opt.res, opt.chrom_file, 
            opt.logscale, opt.pad, opt.std, opt.rp, opt.tol, opt.window_size, opt.step_size,
            opt.output_dist, opt.output_format, opt.mode)
'''



import os
import sys
import time
import numpy as np
import torch
import torch.nn.functional as F
from multiprocessing import Pool
from scipy.sparse import csr_matrix
from scipy.stats import chi2_contingency
from sklearn.decomposition import PCA
from sklearn.cluster import KMeans


def neighbor_ave_gpu(A, pad):
    if pad == 0:
        return torch.from_numpy(A).float().cuda()
    ll = pad * 2 + 1
    conv_filter = torch.ones(1, 1, ll, ll).cuda()
    B = F.conv2d(torch.from_numpy(A[None, None, :, :]).float().cuda(), conv_filter, padding=pad * 2)
    return (B[0, 0, pad:-pad, pad:-pad] / float(ll * ll))


def random_walk_gpu(A, rp):
    ngene, _ = A.shape
    A = A - torch.diag(torch.diag(A))
    A = A + torch.diag(torch.sum(A, 0) == 0).float()
    P = torch.div(A, torch.sum(A, 0))
    Q = torch.eye(ngene).cuda()
    I = torch.eye(ngene).cuda()
    for i in range(30):
        Q_new = (1 - rp) * I + rp * torch.mm(Q, P)
        delta = torch.norm(Q - Q_new, 2)
        Q = Q_new
        if delta < 1e-6:
            break
    return Q


def impute_gpu(args):
    cell, c, ngene, pad, rp = args
    D = np.loadtxt(cell + '_chr' + c + '.txt')
    A = csr_matrix((D[:, 2], (D[:, 0], D[:, 1])), shape=(ngene, ngene)).toarray()
    A = np.log2(A + A.T + 1)
    A = neighbor_ave_gpu(A, pad)
    if rp == -1:
        Q = A[:]
    else:
        Q = random_walk_gpu(A, rp)
    return Q.reshape(ngene * ngene)


def hicluster_gpu(network, chromsize, nc, res=1000000, pad=1, rp=0.5, prct=20, ndim=20):
    matrix = []
    for i, c in enumerate(chromsize):
        ngene = int(chromsize[c] / res) + 1
        start_time = time.time()
        Q_concat = torch.zeros(len(network), ngene * ngene).float().cuda()
        for j, cell in enumerate(network):
            Q_concat[j] = impute_gpu([cell, c, ngene, pad, rp])
        Q_concat = Q_concat.cpu().numpy()
        if prct > -1:
            thres = np.percentile(Q_concat, 100 - prct, axis=1)
        Q_concat = (Q_concat > thres[:, None])
        end_time = time.time()
        print('Load and impute chromosome', c, 'take', end_time - start_time, 'seconds')
        ndim = int(min(Q_concat.shape) * 0.2) - 1
        # U, S, V = torch.svd(Q_concat, some=True)
        # R_reduce = torch.mm(U[:, :ndim], torch.diag(S[:ndim])).cuda().numpy()
        pca = PCA(n_components=ndim)
        R_reduce = pca.fit_transform(Q_concat)
        matrix.append(R_reduce)
        print(c)
    matrix = np.concatenate(matrix, axis=1)
    pca = PCA(n_components=min(matrix.shape) - 1)
    matrix_reduce = pca.fit_transform(matrix)
    kmeans = KMeans(n_clusters=nc, n_init=200).fit(matrix_reduce[:, :ndim])
    return kmeans.labels_, matrix_reduce


def neighbor_ave_cpu(A, pad):
    if pad == 0:
        return A
    ngene, _ = A.shape
    ll = pad * 2 + 1
    B, C, D, E = [np.zeros((ngene + ll, ngene + ll)) for i in range(4)]
    B[(pad + 1):(pad + ngene + 1), (pad + 1):(pad + ngene + 1)] = A[:]
    F = B.cumsum(axis=0).cumsum(axis=1)
    C[ll:, ll:] = F[:-ll, :-ll]
    D[ll:, :] = F[:-ll, :]
    E[:, ll:] = F[:, :-ll]
    return (np.around(F + C - D - E, decimals=8)[ll:, ll:] / float(ll * ll))


def random_walk_cpu(A, rp):
    ngene, _ = A.shape
    A = A - np.diag(np.diag(A))
    A = A + np.diag(np.sum(A, axis=0) == 0)
    P = np.divide(A, np.sum(A, axis=0))
    Q = np.eye(ngene)
    I = np.eye(ngene)
    for i in range(30):
        Q_new = (1 - rp) * I + rp * np.dot(Q, P)
        delta = np.linalg.norm(Q - Q_new)
        Q = Q_new.copy()
        if delta < 1e-6:
            break
    return Q


def impute_cpu(args):
    cell, c, ngene, pad, rp = args
    D = np.loadtxt(cell + '_chr' + c + '.txt')
    A = csr_matrix((D[:, 2], (D[:, 0], D[:, 1])), shape=(ngene, ngene)).toarray()
    A = np.log2(A + A.T + 1)
    A = neighbor_ave_cpu(A, pad)
    if rp == -1:
        Q = A[:]
    else:
        Q = random_walk_cpu(A, rp)
    return [cell, Q.reshape(ngene * ngene)]


def hicluster_cpu(network, chromsize, nc, res=1000000, pad=1, rp=0.5, prct=20, ndim=20, ncpus=10):
    matrix = []
    for i, c in enumerate(chromsize):
        ngene = int(chromsize[c] / res) + 1
        start_time = time.time()
        paras = [[cell, c, ngene, pad, rp] for cell in network]
        p = Pool(ncpus)
        result = p.map(impute_cpu, paras)
        p.close()
        index = {x[0]: j for j, x in enumerate(result)}
        Q_concat = np.array([result[index[x]][1] for x in network])
        if prct > -1:
            thres = np.percentile(Q_concat, 100 - prct, axis=1)
            Q_concat = (Q_concat > thres[:, None])
        end_time = time.time()
        print('Load and impute chromosome', c, 'take', end_time - start_time, 'seconds')
        ndim = int(min(Q_concat.shape) * 0.2) - 1
        pca = PCA(n_components=ndim)
        R_reduce = pca.fit_transform(Q_concat)
        matrix.append(R_reduce)
        print(c)
    matrix = np.concatenate(matrix, axis=1)
    pca = PCA(n_components=min(matrix.shape) - 1)
    matrix_reduce = pca.fit_transform(matrix)
    kmeans = KMeans(n_clusters=nc, n_init=200).fit(matrix_reduce[:, :ndim])
    return kmeans.labels_, matrix_reduce


def raw_pca(network, chromsize, nc, res=1000000, ndim=20):
    matrix = []
    for i, c in enumerate(chromsize):
        ngene = int(chromsize[c] / res) + 1
        start_time = time.time()
        uptri = np.triu_indices(ngene, 1)
        A_concat = np.zeros((len(label), len(uptri[0]))).astype(float)
        j = 0
        for cell in network:
            D = np.loadtxt(cell + '_chr' + c + '.txt')
            A = csr_matrix((D[:, 2], (D[:, 0], D[:, 1])), shape=(ngene, ngene)).toarray()
            A = np.log2(A + A.T + 1)
            A_concat[j, :] = A[uptri]
            j += 1
        end_time = time.time()
        print('Load and impute chromosome', c, 'take', end_time - start_time, 'seconds')
        pca = PCA(n_components=min(A_concat.shape) - 1)
        A_reduce = pca.fit_transform(A_concat)
        matrix.append(A_reduce)
        print(c)
    matrix = np.concatenate(matrix, axis=1)
    pca = PCA(n_components=min(matrix.shape) - 1)
    matrix_reduce = pca.fit_transform(matrix)
    kmeans = KMeans(n_clusters=nc, n_init=200).fit(matrix_reduce[:, 1:(ndim + 1)])
    return kmeans.labels_, matrix_reduce


def ds_pca(network, chromsize, nc, res=1000000, ndim=20):
    matrix = []
    for i, c in enumerate(chromsize):
        ngene = int(chromsize[c] / res) + 1
        start_time = time.time()
        uptri = np.triu_indices(ngene, 1)
        A_concat = np.zeros((len(label), len(uptri[0]))).astype(float)
        j = 0
        tot = int(np.min([np.sum(np.loadtxt(cell + '_chr' + c + '.txt')[:, 2]) for cell in network]))
        for cell in network:
            D = np.loadtxt(cell + '_chr' + c + '.txt')
            A = csr_matrix((ngene, ngene)).toarray().astype(float)
            idx = np.concatenate([[j for i in range(int(D[j, 2]))] for j in range(len(D))])
            shu = np.arange(len(idx))
            np.random.shuffle(shu)
            for x in idx[shu[:tot]]:
                A[int(D[x, 0]), int(D[x, 1])] += 1
            A = np.log2(A + A.T + 1)
            A_concat[j, :] = A[uptri]
            j += 1
        end_time = time.time()
        print('Load and impute chromosome', c, 'take', end_time - start_time, 'seconds')
        pca = PCA(n_components=min(A_concat.shape) - 1)
        A_reduce = pca.fit_transform(A_concat)
        matrix.append(A_reduce)
        print(c)
    matrix = np.concatenate(matrix, axis=1)
    pca = PCA(n_components=min(matrix.shape) - 1)
    matrix_reduce = pca.fit_transform(matrix)
    kmeans = KMeans(n_clusters=nc, n_init=200).fit(matrix_reduce[:, :ndim])
    return kmeans.labels_, matrix_reduce


def comp_comp(O, cg):
    ngene, _ = Q.shape
    N = torch.zeros(ngene, ngene).float().cuda()
    for k in range(ngene):
        diagsum = torch.sum(torch.diag(O, k)) / (ngene - k)
        N = N + torch.diag(torch.diag(O, k) / diagsum, k)
    D = torch.diag(torch.diag(N))
    N = N + torch.t(N) - D
    N[N != N] = 0.0
    C = corrcoef(N).cpu().numpy()
    pca = PCA(n_components=2)
    pc = pca.fit_transform(C)
    corr = np.abs([np.corrcoef(cg, pc[:, i])[0, 1] for i in range(2)])
    # if corr[0] > corr[1]:
    # 	k = 0
    # else:
    # 	k = 1
    k = 0
    pc, corr = pc[:, k], corr[k]
    comp = torch.mean(N[pc > 0][:, pc < 0]).item() * 2 / (
            torch.mean(N[pc > 0][:, pc > 0]).item() + torch.mean(N[pc < 0][:, pc < 0]).item())
    if np.corrcoef(cg, pc)[0, 1] < 0:
        return [C, -pc, corr, comp]
    else:
        return [C, pc, corr, comp]


def compartment(network, chromsize, nc, res=1000000, ndim=20):
    global chrcg
    pca = PCA(n_components=2)
    matrix = []
    for i, c in enumerate(chrom):
        ngene = int(chromsize[c] / res) + 1
        start_time = time.time()
        comp = np.zeros((len(label), ngene)).astype(float)
        j = 0
        for cell in network:
            D = np.loadtxt(cell + '_chr' + c + '.txt')
            A = csr_matrix((D[:, 2], (D[:, 0], D[:, 1])), shape=(ngene, ngene)).toarray()
            A = np.log2(A + A.T + 1)
            A = neighbor_ave_gpu(A, 0)
            comp[j] = comp_comp(A, chrcg[c])
            j += 1
        end_time = time.time()
        print('Load and random walk take', end_time - start_time, 'seconds')
        matrix.append(comp)
        print(c)
    matrix = np.concatenate(matrix, axis=1)
    pca = PCA(n_components=min(matrix.shape) - 1)
    matrix_reduce = pca.fit_transform(matrix)
    kmeans = KMeans(n_clusters=nc, n_init=200).fit(matrix_reduce[:, :ndim])
    return kmeans.labels_, matrix_reduce


def decay(network, chromsize, nc, res=1000000, ndim=20):
    matrix = []
    for i, c in enumerate(chromsize):
        ngene = int(chromsize[c] / res) + 1
        start_time = time.time()
        dec = np.zeros((len(label), ngene - 1)).astype(float)
        j = 0
        for j, cell in enumerate(network):
            D = np.loadtxt(cell + '_chr' + c + '.txt')
            A = csr_matrix((D[:, 2], (D[:, 0], D[:, 1])), shape=(ngene, ngene)).toarray()
            tmp = np.array([np.sum(np.diag(A, k)) for k in range(1, ngene)])
            dec[j] = tmp / np.sum(tmp)
        end_time = time.time()
        print('Load and random walk take', end_time - start_time, 'seconds')
        matrix.append(dec)
        print(c)
    matrix = np.concatenate(matrix, axis=1)
    pca = PCA(n_components=min(matrix.shape) - 1)
    matrix_reduce = pca.fit_transform(matrix)
    kmeans = KMeans(n_clusters=nc, n_init=200).fit(matrix_reduce[:, :ndim])
    return kmeans.labels_, matrix_reduce


def merge_gpu(network, c, res, pad=1, rp=0.5, prct=-1):
    global chromsize
    ngene = int(chromsize[c] / res) + 1
    start_time = time.time()
    Q_sum = torch.zeros(ngene * ngene).float().cuda()
    for cell in network:
        Q_sum = Q_sum + impute_gpu([cell, c, ngene, pad, rp, prct])
    end_time = time.time()
    print('Load and impute chromosome', c, 'take', end_time - start_time, 'seconds')
    Q_sum = Q_sum.cpu().numpy().reshape(ngene, ngene)
    return Q_sum


def merge_cpu(network, c, res, pad=1, rp=0.5, prct=-1):
    global chromsize
    ngene = int(chromsize[c] / res) + 1
    start_time = time.time()
    Q_sum = torch.zeros(ngene * ngene).float().cuda()
    for cell in network:
        Q_sum = Q_sum + impute_cpu([cell, c, ngene, pad, rp, prct])[1]
    end_time = time.time()
    print('Load and impute chromosome', c, 'take', end_time - start_time, 'seconds')
    Q_sum = Q_sum.reshape(ngene, ngene)
    return Q_sum


def output_topdom(cell, c, Q, res):
    global chromsize
    ngene, _ = Q.shape
    B = [['chr' + c, i * res, (i + 1) * res] for i in range(ngene)]
    B[-1][-1] = chromsize[c]
    C = np.concatenate((B, Q), axis=1)
    np.savetxt(cell + '_chr' + c + '.topdommatrix', C, fmt='%s', delimiter='\t')
    return


def output_sparse(cell, c, Q, res):
    idx = np.where(Q > 0)
    C = np.concatenate(((idx[0] * res)[:, None], (idx[1] * res)[:, None], Q[idx][:, None]), axis=1)
    np.savetxt(cell + '_chr' + c + '.sparsematrix', C, fmt='%d\t%d\t%s', delimiter='\n')
    return


def diff_dom(args):
    domlist, cluster, ctlist, res, c, chromsize = args
    start_time = time.time()
    ctdict = {x: i for i, x in enumerate(ctlist)}
    ngene = int(chromsize[c] / res) + 1
    bound = np.zeros((len(cluster), ngene))
    for i, cell in enumerate(domlist):
        domain = np.loadtxt(cell, dtype=np.str)
        C = domain[domain[:, -1] == 'domain', :2].astype(int) / res
        if len(C) > int(chromsize[c] / 10000000) + 1:
            bound[i, np.array(list(set(C.flatten())), dtype=int)] += 1
    cellfilter = (np.sum(bound, axis=1) > 0)
    count = np.array([np.sum(np.logical_and(cluster == k, cellfilter)) for k in ctlist])
    prob = np.array([np.sum(bound[cluster == k], axis=0) for k in ctlist])
    ptmp, btmp = [], []
    for i in range(ngene):
        contig = [[prob[j, i], count[j] - prob[j, i]] for j in range(ctlist)]
        if np.sum(np.sum(contig, axis=0) == 0) == 0:
            ptmp.append(chi2_contingency(contig)[1])
            btmp.append('\t'.join([c, str(i * res), str(i * res + res)]))
    print(c, time.time() - start_time)
    return [bound, prob / count[:, None], btmp, ptmp]


def chrflankpv(args):
    bins, prob, cluster, ctlist, res, c, chromsize, w = args
    start_time = time.time()
    sel = (np.sum(prob, axis=1) > 0)
    tmpclust = cluster[sel]
    prob = prob[sel]
    flankpv = []
    for x in bins:
        x = x.split('\t')
        mid = int(x[1]) // res
        tmpcount = np.max(prob[:, (mid - w):(mid + w + 1)], axis=1)
        contig = np.array(
            [[np.sum(tmpcount[tmpclust == j]), np.sum(tmpclust == j) - np.sum(tmpcount[tmpclust == j])] for j in
             ctlist])
        r = contig[:, 0] / np.sum(contig, axis=1)
        if np.sum(contig == 0) == 0:
            flankpv.append(x + [chi2_contingency(contig)[1]] + r)
    flankpv = np.array(flankpv)
    print(c, time.time() - start_time)
    return [flankpv[:, 4:].astype(float), flankpv[:, 3].astype(float), flankpv[:, :3]]


def filter_bins(prob, fdr, fdr_cutoff=0.05, diff_cutoff=0, max_cutoff=0, min_cutoff=1):
    maxp = np.max(prob, axis=0)
    minp = np.min(prob, axis=0)
    ans = np.logical_and(maxp > max_cutoff, minp < min_cutoff)
    ans = np.logical_and(ans, maxp - minp > diff_cutoff)
    ans = np.logical_and(ans, fdr < fdr_cutoff)
    return ans



# mode=pad2_std1_rp0.5_sqrtvc
# command time python /gale/ddn/snm3C/humanPFC/code/loop_sumcell_chr.py --cell_list /gale/ddn/snm3C/humanPFC/smoothed_matrix/10kb_resolution/filelist/L23_covgroup${i}_${mode}_chr${c}_looplist.txt --outprefix /gale/ddn/snm3C/humanPFC/smoothed_matrix/10kb_resolution/merged/L23_covgroup${i}_${mode}_dist_trim/L23_covgroup${i}_${mode}_dist_trim_chr${c} --res 10000
# command time python /gale/ddn/snm3C/humanPFC/code/loop_sumcell_chr.py --cell_list /gale/ddn/snm3C/humanPFC/smoothed_matrix/10kb_resolution/filelist/L23_${mode}_chr${c}_looplist.txt --group_list /gale/ddn/snm3C/humanPFC/smoothed_matrix/10kb_resolution/filelist/L23_${mode}_chr${c}_grouplist.txt --outprefix /gale/ddn/snm3C/humanPFC/smoothed_matrix/10kb_resolution/merged/L23_${mode}_dist_trim/L23_${mode}_dist_trim_chr${c} --res 10000

import time
import cv2
import h5py
import argparse
import numpy as np
import pandas as pd
from scipy import stats
from scipy.sparse import save_npz, load_npz, csr_matrix, coo_matrix

def loop_sumcell_chr(cell_list, outprefix, res, 
    group_list=None, matrix='QEO', sum_only=False, test_only=False, 
    norm_mode='dist_trim', min_dist=50000, max_dist=10000000, pad=5, gap=2):

    def load_group_hdf(file):
        with h5py.File(file, 'r') as f:
            g = f['Matrix']
            A = csr_matrix((g['data'][()], g['indices'][()], g['indptr'][()]), g.attrs['shape'])
            tmp = g.attrs['count']
            A.data = A.data * tmp
        return [A, tmp]

    def output(file, A):
        f = h5py.File(f'{file}', 'w')
        g = f.create_group('Matrix')
        g.create_dataset('data', data=A.data, dtype='float32', compression='gzip')
        g.create_dataset('indices', data=A.indices, dtype=int, compression='gzip') 
        g.create_dataset('indptr', data=A.indptr, dtype=int, compression='gzip')
        g.attrs['shape'] = A.shape
        g.attrs['count'] = tot
        f.close()

    celllist = np.loadtxt(cell_list, dtype=np.str)
    tot = len(celllist)

    if not test_only:
        thres = stats.norm(0, 1).isf(0.025) 
        with h5py.File(f'{celllist[0]}.hdf5', 'r') as f:
            dim1, dim2 = f['Matrix'].attrs['shape']
        Qsum, Esum, Osum = [csr_matrix((dim1, dim2)) for i in range(3)]
        start_time = time.time()
        if group_list:
            grouplist = np.loadtxt(group_list, dtype=np.str)
            for i, cell in enumerate(grouplist):
                if 'Q' in matrix:
                    A, tmp = load_group_hdf(f'{cell}.hdf5')
                    Qsum += A
                if 'E' in matrix:
                    A, tmp = load_group_hdf(f'{cell}.E.hdf5')
                    Esum += A
                if 'O' in matrix:
                    A, tmp = load_group_hdf(f'{cell}.O.hdf5')
                    Osum += A
                print('Merge', i, 'groups takes', time.time() - start_time, 'seconds')
        else:
            for i, cell in enumerate(celllist):
                if 'Q' in matrix:
                    with h5py.File(f'{cell}.hdf5', 'r') as f:
                        g = f['Matrix']
                        Q = csr_matrix((g['data'][()], g['indices'][()], g['indptr'][()]), g.attrs['shape'])
                    Qsum += Q
                if ('E' in matrix) or ('O' in matrix):
                    E = load_npz(f'{cell}_{norm_mode}.E.npz')
                if 'E' in matrix:
                    Esum += E
                if 'O' in matrix:
                    O = E.copy()
                    O.data = (O.data > thres).astype(int)
                    Osum += O
                print('Merge', i, 'cells takes', time.time() - start_time, 'seconds')
        if 'Q' in matrix:
            Qsum.data = Qsum.data / tot
            output(f'{outprefix}.hdf5', Qsum)
        if 'E' in matrix:
            Esum.data = Esum.data / tot
            output(f'{outprefix}.E.hdf5', Esum)
        if 'O' in matrix:
            Osum.data = Osum.data / tot
            output(f'{outprefix}.O.hdf5', Osum)
        print('Merge cell', time.time() - start_time)

        if (not 'E' in matrix) or (not 'O' in matrix) or sum_only:
            return

        E = Esum.toarray()
        O = Osum.toarray()
        del Qsum, Esum, Osum

    else:
        E, _ = load_group_hdf(f'{outprefix}.E.hdf5')
        O, _ = load_group_hdf(f'{outprefix}.O.hdf5')

    start_time = time.time()
    oefilter = np.logical_and(E > 0, O > 0.1)
    loop = np.where(oefilter)
    distfilter = np.logical_and((loop[1] - loop[0]) > (min_dist / res), (loop[1] - loop[0]) < (max_dist / res))
    loop = (loop[0][distfilter], loop[1][distfilter])

    start_time = time.time()
    eloop = np.zeros((len(celllist), len(loop[0])))
    for i, cell in enumerate(celllist):
        eloop[i] = load_npz(f'{cell}_{norm_mode}.T.npz')[loop].A.ravel()
    print('Load', len(loop[0]), 'loop', time.time() - start_time)

    start_time = time.time()
    pvr = np.array([stats.wilcoxon(xx, alternative='greater')[1] for xx in eloop.T])
    pvt = stats.ttest_1samp(eloop, 0, axis=0)
    pvt[1][pvt[0] > 0] *= 2
    pvt[1][pvt[0] <= 0] = 1
    pvt = pvt[1]
    print('Test loop', time.time() - start_time)
    del eloop

    w = pad * 2 + 1
    start_time = time.time()

    kernel_bl = np.zeros((w, w), np.float32)
    kernel_bl[-pad:, :(pad - gap)] = 1
    kernel_bl[-(pad - gap):, :pad] = 1

    kernel_donut = np.ones((w, w), np.float32)
    kernel_donut[pad, :] = 0
    kernel_donut[:, pad] = 0
    kernel_donut[(pad - gap):(pad + gap + 1), (pad - gap):(pad + gap + 1)] = 0

    kernel_lr = np.ones((3, w), np.float32)
    kernel_lr[:, (pad - gap):(pad + gap + 1)] = 0

    kernel_bu = np.ones((w, 3), np.float32)
    kernel_bu[(pad - gap):(pad + gap + 1), :] = 0

    kernel_bl = kernel_bl / np.sum(kernel_bl)
    kernel_donut = kernel_donut / np.sum(kernel_donut)
    kernel_lr = kernel_lr / np.sum(kernel_lr)
    kernel_bu = kernel_bu / np.sum(kernel_bu)

    Ebl = cv2.filter2D(E, -1, kernel=kernel_bl) * (E > 0)
    Edonut = cv2.filter2D(E, -1, kernel=kernel_donut) * (E > 0)
    Elr = cv2.filter2D(E, -1, kernel=kernel_lr) * (E > 0)
    Ebu = cv2.filter2D(E, -1, kernel=kernel_bu) * (E > 0)

    data = pd.DataFrame(np.array([loop[0], loop[1], pvr, pvt, E[loop], Ebl[loop], Edonut[loop], Elr[loop], Ebu[loop]]).T, columns = ['x1', 'y1', 'rpv', 'tpv', 'E', 'E_bl', 'E_donut', 'E_h', 'E_v'])
    data.to_hdf(f'{outprefix}.loop.hdf5', key='loop')
    print('Bulk bkg', time.time() - start_time)

    return

'''
parser = argparse.ArgumentParser()
parser.add_argument('--cell_list', type=str, default=None, help='Full path of a file containing the full path of all imputed files to be merged without .hdf5 suffix')
parser.add_argument('--outprefix', type=str, default=None, help='Prefix of sumed matrix including directory')
parser.add_argument('--res', type=int, default=None, help='Bin size as integer to generate contact matrix')
parser.add_argument('--group_list', type=str, default=None, help='Full path of a file containing the full path of all summed imputed files to be merged without .hdf5 suffix')
parser.add_argument('--matrix', type=str, default='QEO', help='Types of matrices to sum') #Q E O
parser.add_argument('--sum_only', dest='sum_only', action='store_true', help='Sum cells only and do not test for loops')
parser.set_defaults(sum_only=False)
parser.add_argument('--test_only', dest='test_only', action='store_true', help='Sum of cells already exist and do loop test only')
parser.set_defaults(test_only=False)
parser.add_argument('--norm_mode', type=str, default='dist_trim', help='Suffix of normalized file names')
parser.add_argument('--min_dist', type=int, default=50000, help='Minimum distance threshold of loop')
parser.add_argument('--max_dist', type=int, default=10000000, help='Maximum distance threshold of loop')
parser.add_argument('--pad', type=int, default=5, help='One direction size of larger square for donut background')
parser.add_argument('--gap', type=int, default=2, help='One direction size of smaller square for donut background')
opt = parser.parse_args()

loop_sumcell_chr(opt.cell_list, opt.outprefix, opt.res, 
        opt.group_list, opt.matrix, opt.sum_only, opt.test_only,
        opt.norm_mode, opt.min_dist, opt.max_dist, opt.pad, opt.gap)
'''



import pandas as pd
import numpy as np
import pathlib
import cooler
import h5py
from concurrent.futures import ProcessPoolExecutor, as_completed

from ..cool import get_chrom_offsets
from .merge_cell_to_group import read_single_cool_chrom


def chrom_sum_iterator(cell_urls,
                       chrom_sizes,
                       chrom_offset):
    for chrom in chrom_sizes.keys():
        # sum together multiple chunks
        # first
        cell_url = cell_urls[0]
        matrix = read_single_cool_chrom(cell_url, chrom)
        # others
        for cell_url in cell_urls[1:]:
            matrix += read_single_cool_chrom(cell_url, chrom)

        matrix = matrix.tocoo()
        pixel_df = pd.DataFrame({
            'bin1_id': matrix.row,
            'bin2_id': matrix.col,
            'count': matrix.data
        })
        pixel_df.iloc[:, :2] += chrom_offset[chrom]
        yield pixel_df


def save_single_matrix_type(cooler_path,
                            bins_df,
                            cell_urls,
                            chrom_sizes,
                            chrom_offset):
    chrom_iter = chrom_sum_iterator(cell_urls,
                                    chrom_sizes,
                                    chrom_offset)
    cooler.create_cooler(cool_uri=cooler_path,
                         bins=bins_df,
                         pixels=chrom_iter,
                         ordered=True,
                         dtypes={'count': np.float32})
    with h5py.File(cooler_path, 'a') as f:
        f.attrs['group_n_cells'] = len(cell_urls)
    return


def make_raw_matrix_cell_table(cell_table, resolution_str='10K'):
    # get all the raw matrix cell url automatically if the dir path is the default
    scool_dirs = cell_table['cell_url'].apply(lambda i: '/'.join(i.split('/')[:-4])).unique()
    cell_urls = {}
    scool_file_pattern = f'raw/*.{resolution_str}.scool'
    for scool_dir in scool_dirs:
        for scool_path in pathlib.Path(scool_dir).glob(scool_file_pattern):
            with h5py.File(scool_path, 'r') as _cool:
                cell_ids = list(_cool['cells'].keys())
                for cell_id in cell_ids:
                    cell_urls[cell_id] = f'{scool_path}::/cells/{cell_id}'
    raw_url_series = pd.Series(cell_urls)

    # delete old url
    del cell_table['cell_url']
    # add new url
    cell_table['cell_url'] = raw_url_series

    na_cells = cell_table['cell_url'].isna().sum()
    if na_cells > 0:
        raise ValueError(f'{na_cells} cells do not have raw matrix.')

    cell_table = cell_table[['cell_url', 'cell_group']]
    return cell_table


def merge_raw_scool_by_cluster(chrom_size_path, resolution, cell_table_path,
                               output_dir, cpu=1):
    """Sum the raw matrix of cells, no normalization."""
    # determine chunk dirs for the group:
    output_dir = pathlib.Path(output_dir).absolute()
    output_dir.mkdir(exist_ok=True)
    cell_table = pd.read_csv(cell_table_path,
                             sep='\t',
                             index_col=0,
                             header=None,
                             names=['cell_id', 'cell_url', 'cell_group'])

    chrom_sizes = cooler.read_chromsizes(chrom_size_path, all_names=True)
    bins_df = cooler.binnify(chrom_sizes, resolution)
    chrom_offset = get_chrom_offsets(bins_df)

    with ProcessPoolExecutor(cpu) as exe:
        futures = {}
        for cell_group, sub_df in cell_table.groupby('cell_group'):
            cell_urls = sub_df['cell_url'].tolist()
            cooler_path = str(output_dir / f'{cell_group}.cool')
            future = exe.submit(save_single_matrix_type,
                                cooler_path=cooler_path,
                                bins_df=bins_df,
                                cell_urls=cell_urls,
                                chrom_sizes=chrom_sizes,
                                chrom_offset=chrom_offset)
            futures[future] = cell_group
        for future in as_completed(futures):
            cell_group = futures[future]
            print(f'Matrix {cell_group} generated')
            future.result()
    return



import time
import h5py
import numpy as np
from scipy.sparse import save_npz, csr_matrix, vstack
from sklearn.decomposition import TruncatedSVD


def run_svd(matrix):
    svd = TruncatedSVD(n_components=50, algorithm='arpack')
    matrix_reduce = svd.fit_transform(matrix)
    matrix_reduce = matrix_reduce / svd.singular_values_
    return matrix_reduce


def concat_cell(indir, chrom, mode, res, dist=10000000, save_raw=True, scale_factor=100000):
    # TODO simplify chrom definition
    if chrom[:3] == 'chr':
        chrom = chrom[3:]

    # TODO simplify file IO here
    celllist = np.loadtxt(indir + '../cell_list.txt', dtype=np.str)

    f = h5py.File(indir + 'chr' + chrom + '/' + celllist[0] + '_chr' + chrom + '_' + mode + '.hdf5', 'r')
    ngene = f['Matrix'].attrs['shape'][0]
    f.close()
    idx = np.triu_indices(ngene, k=1)
    idxfilter = np.array([(yy - xx) < (dist / res + 1) for xx, yy in zip(idx[0], idx[1])])
    idx = (idx[0][idxfilter], idx[1][idxfilter])

    start_time = time.time()
    matrix = []
    for i, cell in enumerate(celllist):
        f = h5py.File(indir + 'chr' + chrom + '/' + cell + '_chr' + chrom + '_' + mode + '.hdf5', 'r')
        g = f['Matrix']
        A = csr_matrix((g['data'][()], g['indices'][()], f['indptr'][()]), g.attrs['shape'])
        f.close()
        matrix.append(csr_matrix(A[idx]))
        if i % 100 == 0:
            print(i, 'cells loaded', time.time() - start_time, 'seconds')

    matrix = vstack(matrix)

    if save_raw:
        save_npz(indir + 'merged/' + mode + '_chr' + chrom + '.npz', matrix)

    matrix.data = matrix.data * scale_factor
    matrix_reduce = run_svd(matrix.data)
    np.save(indir + 'merged/' + mode + '_chr' + chrom + '.svd50.npy', matrix_reduce)
    return



"""
CLI defined here

When adding new function:
1. add a func_register_subparser function to register the subparser
2. add a condition in main func about this new func name, import the real func as func in main
"""
import schicluster
from schicluster import __version__

import argparse
import inspect
import subprocess
import sys
import logging
import os

os.environ["NUMEXPR_MAX_THREADS"] = "8"

log = logging.getLogger()

DESCRIPTION = """
scHiCluster is a toolkit for single-cell HiC data preprocessing, imputation, and clustering analysis.

Current Tool List in scHiCluster:

"""

EPILOG = ''


class NiceFormatter(logging.Formatter):
    """
    From Cutadapt https://github.com/marcelm/cutadapt
    Do not prefix "INFO:" to info-level log messages (but do it for all other
    levels).
    Based on http://stackoverflow.com/a/9218261/715090 .
    """

    def format(self, record):
        if record.levelno != logging.INFO:
            record.msg = '{}: {}'.format(record.levelname, record.msg)
        return super().format(record)


def validate_environment():
    try:
        # TODO add env validation here
        subprocess.run(['bedtools', '--version'],
                       stderr=subprocess.PIPE,
                       stdout=subprocess.PIPE,
                       encoding='utf8',
                       check=True)
    except subprocess.CalledProcessError as e:
        print(e.stderr)
        raise e
    return True


def setup_logging(stdout=False, quiet=False, debug=False):
    """
    From Cutadapt https://github.com/marcelm/cutadapt
    Attach handler to the global logger object
    """
    # Due to backwards compatibility, logging output is sent to standard output
    # instead of standard error if the -o option is used.
    stream_handler = logging.StreamHandler(sys.stdout if stdout else sys.stderr)
    stream_handler.setFormatter(NiceFormatter())
    # debug overrides quiet
    if debug:
        level = logging.DEBUG
    elif quiet:
        level = logging.ERROR
    else:
        level = logging.INFO
    stream_handler.setLevel(level)
    log.setLevel(level)
    log.addHandler(stream_handler)


def _str_to_bool(v: str) -> bool:
    if v.lower() in {'true', '1', 't', 'y', 'yes', 'yeah', 'yup', 'certainly', 'uh-huh'}:
        return True
    else:
        return False


def comp_cpg_cell_register_subparser(subparser):
    parser = subparser.add_parser('comp-cpg-cell',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="")

    parser_req = parser.add_argument_group("required arguments")
    parser_req.add_argument('--indir', type=str, default=None,
                            help='Directory of raw/imputed matrices end with /')

    parser_req.add_argument('--outdir', type=str, default=None,
                            help='Directory of compartment matrix end with /')

    parser_req.add_argument('--cell', type=str, default=None,
                            help='Specific identifier of a cell')

    parser_req.add_argument('--chrom', type=str, default=None,
                            help='Chromosome to compute compartment')

    parser.add_argument('--mode', type=str, default=None,
                        help='Suffix of imputed matrix file names or raw')

    parser.add_argument('--cpg_file', type=str, default=None,
                        help='CpG counts generated by bedtools nuc')


def comp_concatcell_chr_register_subparser(subparser):
    parser = subparser.add_parser('comp-concatcell-chr',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="")

    parser_req = parser.add_argument_group("required arguments")
    parser_req.add_argument('--cell_list', type=str, default=None,
                            help='Full path of a file containing the full path of all CpG npy files to be concatenate')

    parser_req.add_argument('--outprefix', type=str, default=None,
                            help='Prefix of concatenated matrix including directory')

    parser.add_argument('--ncpus', type=int, default=10,
                        help='# threads for parallelization')


def domain_insulation_cell_register_subparser(subparser):
    parser = subparser.add_parser('domain-insulation-cell',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="")

    parser_req = parser.add_argument_group("required arguments")
    parser_req.add_argument('--indir', type=str, default=None,
                            help='Directory of imputed matrices end with /')

    parser_req.add_argument('--cell', type=str, default=None,
                            help='Specific identifier of a cell')

    parser_req.add_argument('--chrom', type=str, default=None,
                            help='Chromosome to compute insulation score')

    parser_req.add_argument('--mode', type=str, default=None,
                            help='Suffix of imputed matrix file names')

    parser.add_argument('--w', type=int, default=10,
                        help='Window size for insulation score')


def domain_concatcell_chr_register_subparser(subparser):
    parser = subparser.add_parser('domain-concatcell-chr',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="")

    parser_req = parser.add_argument_group("required arguments")
    parser_req.add_argument('--cell_list', type=str, default=None,
                            help='Full path of a file containing the full path of all insulation npy or domain txt files to be concatenate')

    parser_req.add_argument('--outprefix', type=str, default=None,
                            help='Prefix of concatenated matrix including directory')

    parser_req.add_argument('--res', type=int, default=None,
                            help='Bin size as integer')

    parser.add_argument('--input_type', type=str, default='insulation',
                        help='Whether input files are insulation.npy or domain.txt')  # insulation or boundary

    parser.add_argument('--ncpus', type=int, default=10,
                        help='# threads for parallelization')


def embed_concatcell_chr_register_subparser(subparser):
    parser = subparser.add_parser('embed-concatcell-chr',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="")

    parser_req = parser.add_argument_group("required arguments")
    parser_req.add_argument('--cell_list', type=str, default=None,
                            help='Full path of a file containing the full path to all imputed matrices to be concatenate')

    parser_req.add_argument('--outprefix', type=str, default=None,
                            help='Prefix of concatenated matrix including directory')

    parser_req.add_argument('--res', type=int, default=None,
                            help='Bin size as integer of imputed matrices')

    parser.add_argument('--dist', type=int, default=10000000,
                        help='Maximum distance threshold of contacts to use')

    parser.add_argument('--skip_raw', dest='save_raw', action='store_false',
                        help='Not to save cell-by-feature matrix before SVD')
    parser.set_defaults(save_raw=True)

    parser.add_argument('--dim', type=int, default=50,
                        help='Number of dimensions to return from SVD')


def embed_mergechr_register_subparser(subparser):
    parser = subparser.add_parser('embed-mergechr',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="")

    parser_req = parser.add_argument_group("required arguments")
    parser_req.add_argument('--embed_list', type=str, default=None,
                            help='Full path of a file containing the full path to dimension reduction files of all chromosomes')

    parser_req.add_argument('--outprefix', type=str, default=None,
                            help='Prefix of final dimension reduction file including directory')

    parser.add_argument('--dim', type=int, default=20,
                        help='Number of dimensions to return from SVD')

    parser.add_argument('--use_pc', dest='norm_sig', action='store_false',
                        help='Not to normalize PCs by singular values')
    parser.set_defaults(norm_sig=True)


def generatematrix_cell_register_subparser(subparser):
    parser = subparser.add_parser('generatematrix-cell',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help='')

    parser_req = parser.add_argument_group('required arguments')
    parser_req.add_argument('--infile', type=str, default=None,
                            help='Path to the short format contact file')

    parser_req.add_argument('--outdir', type=str, default=None,
                            help='Output directory end with /')

    parser_req.add_argument('--cell', type=str, default=None,
                            help='Specific identifier of a cell')

    parser_req.add_argument('--res', type=int, default=None,
                            help='Bin size as integer to generate contact matrix')

    parser_req.add_argument('--chrom_file', type=str, default=None,
                            help='Path to the chromosome size files containing all chromosome to be analyzed as the first column')

    parser.add_argument('--split_file', type=str, default=None,
                        help='Path to the bed file containing all chromosomes need to split and one region per chromosome as splitting point')

    parser.add_argument('--chr1', type=int, default=1,
                        help='Column of fragment1 chromosome, counting starts from 0')

    parser.add_argument('--pos1', type=int, default=2,
                        help='Column of fragment1 position, counting starts from 0')

    parser.add_argument('--chr2', type=int, default=5,
                        help='Column of fragment2 chromosome, counting starts from 0')

    parser.add_argument('--pos2', type=int, default=6,
                        help='Column of fragment2 position, counting starts from 0')

    parser.add_argument('--dist', type=int, default=2500,
                        help='Minimum distance threshold of contacts to use')


def impute_cell_register_subparser(subparser):
    parser = subparser.add_parser('impute-cell',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="")

    parser_req = parser.add_argument_group("required arguments")
    parser_req.add_argument('--indir', type=str, default=None,
                            help='Directory of the contact matrix')

    parser_req.add_argument('--outdir', type=str, default=None,
                            help='Output directory end with /')

    parser_req.add_argument('--cell', type=str, default=None,
                            help='Specific identifier of a cell')

    parser_req.add_argument('--chrom', type=str, default=None,
                            help='Chromosome to impute')

    parser_req.add_argument('--res', type=int, default=None,
                            help='Bin size as integer to generate contact matrix')

    parser_req.add_argument('--chrom_file', type=str, default=None,
                            help='Path to the chromosome size files containing all chromosomes to be analyzed')

    parser.add_argument('--logscale', dest='logscale', action='store_true',
                        help='Log transform raw count')
    parser.set_defaults(logscale=False)

    parser.add_argument('--pad', type=int, default=1,
                        help='Gaussian kernal size')

    parser.add_argument('--std', type=float, default=1,
                        help='Gaussian kernal standard deviation')

    parser.add_argument('--rp', type=float, default=0.5,
                        help='Restart probability of RWR')

    parser.add_argument('--tol', type=float, default=0.01,
                        help='Convergence tolerance of RWR')

    parser.add_argument('--window_size', type=int, default=500000000,
                        help='Size of RWR sliding window')

    parser.add_argument('--step_size', type=int, default=10000000,
                        help='Step length of RWR sliding window')

    parser.add_argument('--output_dist', type=int, default=500000000,
                        help='Maximum distance threshold of contacts when writing output file')

    parser.add_argument('--output_format', type=str, default='hdf5',
                        help='Output file format (hdf5 or npz)')

    parser.add_argument('--mode', type=str, default=None,
                        help='Suffix of output file name')


def loop_bkg_cell_register_subparser(subparser):
    parser = subparser.add_parser('loop-bkg-cell',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="")

    parser_req = parser.add_argument_group("required arguments")
    parser_req.add_argument('--indir', type=str, default=None,
                            help='Directory of imputed matrix')

    parser_req.add_argument('--cell', type=str, default=None,
                            help='Specific identifier of a cell')

    parser_req.add_argument('--chrom', type=str, default=None,
                            help='Chromosome to compute background')

    parser_req.add_argument('--impute_mode', type=str, default=None,
                            help='Suffix of imputed matrix file names')

    parser_req.add_argument('--res', type=int, default=None,
                            help='Bin size as integer to generate contact matrix')

    parser.add_argument('--dist', type=int, default=10050000,
                        help='Maximum distance threshold of contacts to use')

    parser.add_argument('--cap', type=int, default=5,
                        help='Trim Z-scores over the threshold')

    parser.add_argument('--pad', type=int, default=5,
                        help='One direction size of larger square for donut background')

    parser.add_argument('--gap', type=int, default=2,
                        help='One direction size of smaller square for donut background')

    parser.add_argument('--norm_mode', type=str, default='dist_trim',
                        help='Suffix of normalized file names')


def loop_sumcell_chr_register_subparser(subparser):
    parser = subparser.add_parser('loop-sumcell-chr',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="")

    parser_req = parser.add_argument_group("required arguments")
    parser_req.add_argument('--cell_list', type=str, default=None,
                            help='Full path of a file containing the full path of all imputed files to be summed without .hdf5 suffix')

    parser_req.add_argument('--outprefix', type=str, default=None,
                            help='Prefix of summed matrix including directory')

    parser_req.add_argument('--res', type=int, default=None,
                            help='Bin size as integer of imputed matrices')

    parser.add_argument('--group_list', type=str, default=None,
                        help='Full path of a file containing the full path of all summed imputed files to be merged without .hdf5 suffix')

    parser.add_argument('--matrix', type=str, default='QEO',
                        help='Types of matrices to merge (Q: imputed matrix; E: OVE matrix; O: Outlier matrix)')

    parser.add_argument('--sum_only', dest='sum_only', action='store_true',
                        help='Sum cells only and do not test for loops')
    parser.set_defaults(sum_only=False)

    parser.add_argument('--test_only', dest='test_only', action='store_true',
                        help='Sum of cells already exist and do loop test only')
    parser.set_defaults(test_only=False)

    parser.add_argument('--norm_mode', type=str, default='dist_trim',
                        help='Suffix of normalized file names')

    parser.add_argument('--min_dist', type=int, default=50000,
                        help='Minimum distance threshold of loop')

    parser.add_argument('--max_dist', type=int, default=10000000,
                        help='Maximum distance threshold of loop')

    parser.add_argument('--pad', type=int, default=5,
                        help='One direction size of larger square for donut background')

    parser.add_argument('--gap', type=int, default=2,
                        help='One direction size of smaller square for donut background')


def loop_mergechr_register_subparser(subparser):
    parser = subparser.add_parser('loop-mergechr',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="")

    parser_req = parser.add_argument_group("required arguments")
    parser_req.add_argument('--inprefix', type=str, default=None,
                            help='Prefix of summed matrix including directory')

    parser_req.add_argument('--outprefix', type=str, default=None,
                            help='Prefix of loop files including directory')

    parser_req.add_argument('--chrom_file', type=str, default=None,
                            help='Path to the chromosome size files containing all chromosome to be analyzed as the first column')

    parser.add_argument('--split_file', type=str, default=None,
                        help='Path to the bed file containing all chromosomes need to split and one region per chromosome as splitting point')

    parser.add_argument('--res', type=int, default=10000,
                        help='Bin size as integer of loop calling')

    parser.add_argument('--thres_bl', type=int, default=1.33,
                        help='Lowest fold change threshold against bottom left background')

    parser.add_argument('--thres_d', type=int, default=1.33,
                        help='Lowest fold change threshold against donut background')

    parser.add_argument('--thres_h', type=int, default=1.2,
                        help='Lowest fold change threshold against horizontal background')

    parser.add_argument('--thres_v', type=int, default=1.2,
                        help='Lowest fold change threshold against vertical background')

    parser.add_argument('--fdr_thres', type=int, default=0.1,
                        help='Highest t-test FDR threshold of loops')

    parser.add_argument('--dist_thres', type=int, default=20000,
                        help='Highest distance threshold to merge loops into summit')

    parser.add_argument('--size_thres', type=int, default=1,
                        help='Lowest loop number threshold of summit')


def generate_scool_register_subparser(subparser):
    parser = subparser.add_parser('generate-scool',
                                  aliases=['scool'],
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="")

    parser_req = parser.add_argument_group("required arguments")
    parser_req.add_argument('--contacts_table', type=str, required=True, default=None,
                            help='tab-separated table containing tow columns, 1) cell id, '
                                 '2) cell contact file path (juicer-pre format). No header')
    parser_req.add_argument('--output_prefix', type=str, required=True, default=None,
                            help='Output prefix of the scool files. '
                                 'Output path will be {output_prefix}.{resolution_str}.scool')
    parser_req.add_argument('--chrom_size_path', type=str, required=True, default=None,
                            help='Path to the chromosome size file, this file should be in UCSC chrom.sizes format. '
                                 'We will use this file as the reference to create matrix. '
                                 'It is recommended to remove small contigs or chrM from this file.')
    parser_req.add_argument('--resolutions', type=int, nargs='+', required=True, default=[],
                            help='Resolutions to generate the matrix. '
                                 'You can provide multiple resolutions separated by space, '
                                 '(e.g., "--resolutions 10000 100000"). '
                                 'Each resolution will be stored in a separate file.')
    parser.add_argument('--blacklist_1d_path', type=str, required=False, default=None,
                        help='Path to blacklist region BED file, such as ENCODE blacklist. '
                             'Either side of the contact overlapping with a blacklist region will be removed.')
    parser.add_argument('--blacklist_2d_path', type=str, required=False, default=None,
                        help='Path to blacklist region pair BEDPE file. '
                             'Both side of the contact overlapping with the same '
                             'blacklist region pair will be removed.')
    parser.add_argument('--blacklist_resolution', type=int, default=10000, required=False,
                        help='Resolution in bps when consider the 2D blacklist region pairs.')
    parser.add_argument('--not_remove_duplicates', dest='remove_duplicates', action='store_false',
                        help='If set, will NOT remove duplicated contacts based on '
                             '[chr1, pos1, chr2, pos2] values')
    parser.set_defaults(remove_duplicates=True)
    parser.add_argument('--chr1', type=int, default=1, help='0 based index of chr1 column.')
    parser.add_argument('--chr2', type=int, default=5, help='0 based index of chr2 column.')
    parser.add_argument('--pos1', type=int, default=2, help='0 based index of pos1 column.')
    parser.add_argument('--pos2', type=int, default=6, help='0 based index of pos2 column.')
    parser.add_argument('--min_pos_dist', type=int, default=2500,
                        help='Minimum distance for a fragment to be considered.')
    parser.add_argument('--cpu', type=int, default=1, help='number of cpus to parallel.')
    parser.add_argument('--batch_n', type=int, default=50, help='number of cells to deal with in each cpu process.')


def prepare_imputation_register_subparser(subparser):
    parser = subparser.add_parser('imputation',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="")

    parser_req = parser.add_argument_group("required arguments")
    parser_req.add_argument('--input_scool', type=str, required=True, default=None,
                            help='Path to input scool file')
    parser_req.add_argument('--output_dir', type=str, required=True, default=None,
                            help='Path to output directory')
    parser_req.add_argument('--chrom_size_path', type=str, required=True, default=None,
                            help='Path to chromosome sizes file')
    parser_req.add_argument('--output_dist', type=int, required=True, default=None,
                            help='')
    parser_req.add_argument('--window_size', type=int, required=True, default=None,
                            help='')
    parser_req.add_argument('--step_size', type=int, required=True, default=None,
                            help='')
    parser_req.add_argument('--resolution', type=int, required=True, default=None,
                            help='')

    parser.add_argument('--batch_size', type=int, required=False, default=100,
                        help='')
    parser.add_argument('--logscale', dest='logscale', action='store_true',
                        help='')
    parser.set_defaults(logscale=False)
    parser.add_argument('--pad', type=int, required=False, default=1,
                        help='')
    parser.add_argument('--std', type=int, required=False, default=1,
                        help='')
    parser.add_argument('--rp', type=float, required=False, default=0.5,
                        help='')
    parser.add_argument('--tol', type=float, required=False, default=0.01,
                        help='')
    parser.add_argument('--min_cutoff', type=float, required=False, default=1e-5,
                        help='')
    parser.add_argument('--cpu_per_job', type=int, required=False, default=10,
                        help='')
    return


def call_domain_register_subparser(subparser):
    parser = subparser.add_parser('domain',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="")

    parser_req = parser.add_argument_group("required arguments")
    parser_req.add_argument('--cell_table_path', type=str, required=True, default=None,
                            help='Path to cell table file')
    parser_req.add_argument('--output_prefix', type=str, required=True, default=None,
                            help='Output files prefix. '
                                 'The domain boundary matrix will be saved as '
                                 '{output_prefix}.boundary.h5ad (anndata.AnnData); '
                                 'The insulation score matrix will be saved as '
                                 '{output_prefix}.insulation.nc (xarray.DataSet)')
    parser.add_argument('--resolution', type=int, required=False, default=25000,
                        help='Matrix resolution')
    parser.add_argument('--window_size', type=int, required=False, default=10,
                        help='Window size for calculating insulation score')
    parser.add_argument('--cpu', type=int, required=False, default=10,
                        help='Number of CPUs to use')


def call_compartment_register_subparser(subparser):
    parser = subparser.add_parser('compartment',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="")

    parser_req = parser.add_argument_group("required arguments")
    parser_req.add_argument('--cell_table_path', type=str, required=True, default=None,
                            help='Path to cell table file')
    parser_req.add_argument('--output_prefix', type=str, required=True, default=None,
                            help='Output files prefix. '
                                 'The compartment score matrix will be saved as '
                                 '{output_prefix}.compartment.h5ad (anndata.AnnData).')
    parser_req.add_argument('--cpg_profile_path', type=str, required=True, default=None,
                            help='Genome bins CpG ratio. Use "schicluster cpg-ratio" to calculate')
    parser.add_argument('--cpu', type=int, required=False, default=10,
                        help='Number of CPUs to use')
    parser.add_argument('--calc_strength', dest='calc_strength', action='store_true',
                        help='Calculate compartment strength summary')
    parser.set_defaults(calc_strength=False)
    return


def cpg_ratio_register_subparser(subparser):
    parser = subparser.add_parser('cpg-ratio',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="")

    parser_req = parser.add_argument_group("required arguments")
    parser_req.add_argument('--cell_url', type=str, required=True, default=None,
                            help='Path to a cell Cooler URL')
    parser_req.add_argument('--fasta_path', type=str, required=True, default=None,
                            help='Path to genome FASTA file')
    parser_req.add_argument('--hdf_output_path', type=str, required=True, default=None,
                            help='Output path of the CpG ratio')


def embedding_register_subparser(subparser):
    parser = subparser.add_parser('embedding',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="")

    parser_req = parser.add_argument_group("required arguments")
    parser_req.add_argument('--cell_table_path', type=str, required=True, default=None,
                            help='Path to cell table file')
    parser_req.add_argument('--output_dir', type=str, required=True, default=None,
                            help='Path to embedding output dir')
    parser.add_argument('--dim', type=int, required=False, default=50,
                        help='')
    parser.add_argument('--dist', type=int, required=False, default=1000000,
                        help='')
    parser.add_argument('--resolution', type=int, required=False, default=100000,
                        help='')
    parser.add_argument('--scale_factor', type=int, required=False, default=100000,
                        help='')
    parser.add_argument('--cpu', type=int, required=False, default=1,
                        help='')
    parser.add_argument('--norm_sig', dest='norm_sig', action='store_true',
                        help='')
    parser.set_defaults(norm_sig=False)
    parser.add_argument('--save_model', dest='save_model', action='store_true',
                        help='')
    parser.set_defaults(save_model=False)
    parser.add_argument('--save_raw', dest='save_raw', action='store_true',
                        help='')
    parser.set_defaults(save_raw=False)


def main():
    parser = argparse.ArgumentParser(description=DESCRIPTION,
                                     epilog=EPILOG,
                                     formatter_class=argparse.RawDescriptionHelpFormatter,
                                     )
    subparsers = parser.add_subparsers(
        title="functions",
        dest="command",
        metavar=""
    )

    # add subparsers
    current_module = sys.modules[__name__]
    # get all functions in parser
    for name, register_subparser_func in inspect.getmembers(current_module, inspect.isfunction):
        if 'register_subparser' in name:
            register_subparser_func(subparsers)

    # initiate
    args = None
    if len(sys.argv) > 1:
        # print out version
        if sys.argv[1] in ['-v', '--version']:
            print(schicluster.__version__)
            exit()
        else:
            args = parser.parse_args()
    else:
        # print out help
        parser.parse_args(["-h"])
        exit()

    # set up logging
    if not logging.root.handlers:
        setup_logging(stdout=True,
                      quiet=False)
    # execute command
    args_vars = vars(args)
    for k, v in args_vars.items():
        log.info(f'{k}\t{v}')

    cur_command = args_vars.pop('command').lower().replace('_', '-')
    # Do real import here:
    if cur_command in ['generatematrix-cell']:
        from .draft.generatematrix_cell import generatematrix_cell as func
    elif cur_command in ['impute-cell']:
        from .draft.impute_cell import impute_cell as func
    elif cur_command in ['embed-concatcell-chr']:
        from .draft.embed_concatcell_chr import embed_concatcell_chr as func
    elif cur_command in ['embed-mergechr']:
        from .draft.embed_mergechr import embed_mergechr as func
    elif cur_command in ['loop-bkg-cell']:
        from .draft.loop_bkg_cell import loop_bkg_cell as func
    elif cur_command in ['loop-sumcell-chr']:
        from .draft.loop_sumcell_chr import loop_sumcell_chr as func
    elif cur_command in ['loop-mergechr']:
        from .draft.loop_mergechr import loop_mergechr as func
    elif cur_command in ['domain-insulation-cell']:
        from .draft.domain_insulation_cell import domain_insulation_cell as func
    elif cur_command in ['domain-concatcell-chr']:
        from .draft.domain_concatcell_chr import domain_concatcell_chr as func
    elif cur_command in ['comp-cpg-cell']:
        from .draft.comp_cpg_cell import comp_cpg_cell as func
    elif cur_command in ['comp-concatcell-chr']:
        from .draft.comp_concatcell_chr import comp_concatcell_chr as func
    elif cur_command in ['generate-scool', 'scool']:
        from .cool import generate_scool as func
    elif cur_command in ['imputation']:
        from .impute import prepare_impute as func
    elif cur_command in ['domain']:
        from .domain import multiple_call_domain_and_insulation as func
    elif cur_command in ['compartment']:
        from .compartment import multiple_cell_compartment as func
    elif cur_command in ['cpg-ratio']:
        from .compartment import get_cpg_profile as func
    elif cur_command in ['embedding']:
        from .embedding import embedding as func
    else:
        log.debug(f'{cur_command} is not an valid sub-command')
        parser.parse_args(["-h"])
        return

    # run the command
    log.info(f"hicluster: Executing {cur_command}...")
    func(**args_vars)
    log.info(f"{cur_command} finished.")
    return


if __name__ == '__main__':
    main()
